import unittest

import numpy as np

from pyqsp.completion import (CompletionError, cheb2poly,
                              completion_from_root_finding, poly2cheb)
from pyqsp.LPoly import LPoly, PolynomialToLaurentForm

# -----------------------------------------------------------------------------
# unit tests


class Test_completion(unittest.TestCase):
    def test_completion_f_1(self):
        """
        See if completion_from_root_finding can get the G polynomial,
            given the F polynomial in Laurent form
        Test case: F = w should produce G = 0
        """
        lpcoefs = [0, 0, 1]
        Plp = LPoly(lpcoefs, -len(lpcoefs) + 1)

        eps = 1.0e-4
        suc = 1 - 1.0e-4
        p = Plp

        # Capitalization: eps/2 amount of error budget is put to the highest
        # power for sake of numerical stability.
        p_new = suc * \
            (p + LPoly([eps / 4], p.degree) + LPoly([eps / 4], -p.degree))

        # print(f"Laurent F poly {Plp}")
        Qalg = completion_from_root_finding(p_new.coefs, coef_type="F")
        Qlp = Qalg.XPoly
        # print(f"Laurent G poly {Qlp}")

        ncoefs = (p_new * ~p_new + Qlp * ~Qlp).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_f_2(self):
        """
        See if completion_from_root_finding can get the G polynomial,
            given the F polynomial in Laurent form
        Test case: F = -1 + 2 a^2 should produce G = i * 2 a sqrt(1-a^2),
            which has Laurent poly (w^2 - w^(-2))/2
        """
        pcoefs = [-1, 0, 2]
        Plp = PolynomialToLaurentForm(pcoefs)

        eps = 1.0e-4
        suc = 1 - 1.0e-4
        p = Plp

        # Capitalization: eps/2 amount of error budget is put to the highest
        # power for sake of numerical stability.
        p_new = suc * \
            (p + LPoly([eps / 4], p.degree) + LPoly([eps / 4], -p.degree))

        # print(f"Laurent F poly {Plp}")
        Qalg = completion_from_root_finding(p_new.coefs, coef_type="F")
        Qlp = Qalg.XPoly
        # print(f"Laurent G poly {Qlp}")

        self.assertEqual(Qlp.dmin, -2)

        ncoefs = (p_new * ~p_new + Qlp * ~Qlp).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_f_3(self):
        """
        See if completion_from_root_finding can get the G polynomial, given the F polynomial in Laurent form
        Test case: F = a(-3+4a^2) = -3a + 4a^3 = [0, -3, 0, 4] should produce G = (-1 + 4a^2) * sqrt(1-a^2)
        Get:
            Laurent F poly 0.5 * w ^ (-3) + 0.0 * w ^ (-1) + 0.0 * w ^ (1) + 0.5 * w ^ (3)
            Laurent G poly -0.4949998125160129 * w ^ (-3) + 8.24174368309573e-15 * w ^ (-1) + 7.204517437263955e-15 * w ^ (1) + 0.5050001874839872 * w ^ (3)

        cosine matching up with sine -- good
        """
        pcoefs = [0, -3, 0, 4]
        Plp = PolynomialToLaurentForm(pcoefs)

        eps = 1.0e-4
        suc = 1 - 1.0e-4
        p = Plp

        # Capitalization: eps/2 amount of error budget is put to the highest
        # power for sake of numerical stability.
        p_new = suc * \
            (p + LPoly([eps / 4], p.degree) + LPoly([eps / 4], -p.degree))

        # print(f"Laurent F poly {Plp}")
        Qalg = completion_from_root_finding(p_new.coefs, coef_type="F")
        Qlp = Qalg.XPoly
        # print(f"Laurent G poly {Qlp}")

        self.assertEqual(Qlp.dmin, -3)

        ncoefs = (p_new * ~p_new + Qlp * ~Qlp).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_p_1(self):
        pcoefs = [0., -2 + 1j, 0., 2.]

        alg = completion_from_root_finding(pcoefs, coef_type="P")

        F, G = alg.IPoly, alg.XPoly

        ncoefs = (F * ~F + G * ~G).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_p_2(self):
        pcoefs = [-1., 0., 50., 0., -400., 0., 1120., 0., -1280., 0., 512.]

        alg = completion_from_root_finding(pcoefs, coef_type="P")

        F, G = alg.IPoly, alg.XPoly

        ncoefs = (F * ~F + G * ~G).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_p_3(self):
        pcoefs = [-1., 0., (1 / 2) * (4 + 3j - (1 - 2j) * np.sqrt(3)),
                  0., (1 - 1j) * (-1j + np.sqrt(3))]

        alg = completion_from_root_finding(pcoefs, coef_type="P")

        F, G = alg.IPoly, alg.XPoly

        ncoefs = (F * ~F + G * ~G).coefs
        ncoefs_expected = np.zeros(ncoefs.size)
        ncoefs_expected[ncoefs.size // 2] = 1.

        self.assertAlmostEqual(np.max(np.abs(ncoefs - ncoefs_expected)), 0.)

    def test_completion_p_4(self):
        pcoefs = [0., 0., -3., 0., 4.]

        with self.assertRaises(CompletionError):
            completion_from_root_finding(pcoefs, coef_type="P")

    def test_cheb2poly_1(self):
        pcoefs = np.array([0., 0., 1., 2., 2., 0.])
        expected = np.array([1., -6., -14., 8., 16., 0.])
        result = cheb2poly(pcoefs, kind='T')
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)

    def test_cheby2poly_2(self):
        pcoefs = np.array([0., 0., 1., 2., 2., 0.])
        expected = np.array([1., -8., -20., 16., 32., 0.])
        result = cheb2poly(pcoefs, kind='U')
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)

    def test_poly2cheb_1(self):
        pcoefs = np.array([1., -6., -14., 8., 16., 0.])
        expected = np.array([0., 0., 1., 2., 2., 0.])
        result = poly2cheb(pcoefs, kind='T')
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)

    def test_poly2cheb_2(self):
        pcoefs = np.array([1., -8., -20., 16., 32., 0.])
        expected = np.array([0., 0., 1., 2., 2., 0.])
        result = poly2cheb(pcoefs, kind='U')
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)



import os
import unittest

import numpy as np
import tensorflow as tf

from pyqsp import qsp_models

class Test_qsp_models(unittest.TestCase):
    '''
    Unit tests for qsp_models component of pyqsp
    '''
    def setUp(self):
        '''
        Do imports here, so that this module can be entirely optional
        '''
        if self.is_enabled():
            pass
        else:
            print(
                "[pyqsp.test] Skipping qsp_model tests: export PYQSP_TEST_QSP_MODELS=1 to enable these tests")

    def is_enabled(self):
        enabled = 'PYQSP_TEST_QSP_MODELS' in os.environ
        return enabled

    def xtest_hsim1(self):
        if not self.is_enabled():
            return
        t = 3
        poly_deg = 6
        def f(x): return np.cos(t * x)
        model = qsp_models.construct_qsp_model(poly_deg)

        # The intput theta training values
        th_in = np.arange(0, np.pi, np.pi / 50)
        th_in = tf.reshape(th_in, (th_in.shape[0], 1))

        # The desired real part of p(x) which is the upper left value in the unitary of the qsp sequence
        # We also want that Re[q(x)] = 0
        expected_outputs = [f(np.cos(th_in)), np.zeros(th_in.shape)]

        model = qsp_models.construct_qsp_model(poly_deg)
        history = model.fit(x=th_in, y=expected_outputs,
                            epochs=1000, verbose=0)
        # plot_loss(history)
        # plot_qsp_response(f, model)
        response, all_th, circuit_px, circuit_qx = qsp_models.compute_qsp_response(
            model, return_all=True)
        desired = f(np.cos(all_th))
        assert abs(response - desired).mean() < 0.1

    def xtest_mpinverse(self):
        if not self.is_enabled():
            return

        d = 1 / 16
        e = 1 / 16
        k = 2 / d
        poly_deg = int((np.log(1 / e) / d))
        b = np.ceil(k * k * np.log(k / e))
        # odd polynomials for now
        poly_deg = poly_deg if (np.mod(poly_deg, 2) == 1) else (poly_deg + 1)

        # approximation to inverse function d/2x
        def f(x): return np.where(x != 0, d /
                                  2 * (1 - (1 - x ** 2) ** b) / x, 0)

        # The intput theta training values
        th_in = np.arange(0, np.pi, np.pi / 30)
        th_in = tf.reshape(th_in, (th_in.shape[0], 1))

        # The desired real part of p(x) which is the upper left value in the
        # unitary of the qsp sequence
        expected_outputs = [f(np.cos(th_in)), np.zeros(th_in.shape[0])]
        model = qsp_models.construct_qsp_model(poly_deg)
        history = model.fit(x=th_in, y=expected_outputs,
                            epochs=5000, verbose=0)
        # plot_loss(history)
        # plot_qsp_response(f, model)
        response, all_th, circuit_px, circuit_qx = qsp_models.compute_qsp_response(
            model, return_all=True)
        desired = f(np.cos(all_th))
        assert abs(response - desired).mean() < 0.1

    def test_ampamp1(self):
        if not self.is_enabled():
            return
        poly_deg = 19

        # approximation to inverse function d/2x
        def f(x): return np.where(x < 0, -1, np.where(x > 0, 1, 0))

        # The intput theta training values
        th_in = np.arange(0, np.pi, np.pi / 30)
        th_in = tf.reshape(th_in, (th_in.shape[0], 1))

        # The desired real part of p(x) which is the upper left value in the
        # unitary of the qsp sequence
        expected_outputs = [f(np.cos(th_in)), np.zeros(th_in.shape[0])]

        model = qsp_models.construct_qsp_model(poly_deg)
        history = model.fit(x=th_in, y=expected_outputs,
                            epochs=5000, verbose=0)
        response, all_th, circuit_px, circuit_qx = qsp_models.compute_qsp_response(
            model, return_all=True)
        desired = f(np.cos(all_th))
        assert abs(response - desired).mean() < 0.1



import time

import numpy as np
from numpy.polynomial.polynomial import Polynomial

from pyqsp.completion import completion_from_root_finding
from pyqsp.decomposition import angseq
from pyqsp.LPoly import LAlg, LPoly
from pyqsp.poly import StringPolynomial, TargetPolynomial
from pyqsp.response import ComputeQSPResponse


class AngleFindingError(Exception):
    """Raised when angle finding step failes."""
    pass


def angle_sequence(p, eps=1e-4, suc=1 - 1e-4):
    """
    Solve for the angle sequence corresponding to the array p, with eps error budget and suc success probability.
    The bigger the error budget and the smaller the success probability, the better the numerical stability of the process.
    The array p specifies the coefficients of a Laurent polynomial, as powers of w: [ w^(-n), ..., w^(-2), w^(-1), const, w^1, w^2, ..., w^n ]
    This polynomial specifies the 0,0 component of the unitary.
    The QSP phases returned are for the Wz QSP convention (signal unitaries are Z-rotations and QSP phases are X-rotations).
    """
    p = LPoly(p, -len(p) + 1)

    # Capitalization: eps/2 amount of error budget is put to the highest power
    # for sake of numerical stability.
    p_new = suc * \
        (p + LPoly([eps / 4], p.degree) +
         LPoly([eps / 4], -p.degree))

    # Completion phase
    t = time.time()
    g = completion_from_root_finding(p_new.coefs)
    t_comp = time.time()
    print("Completion part finished within time ", t_comp - t)

    # Decomposition phase
    seq = angseq(g)
    t_dec = time.time()
    print("Decomposition part finished within time ", t_dec - t_comp)
    # print(seq)

    # Make sure that the reconstructed element lies in the desired error
    # tolerance regime
    g_recon = LAlg.unitary_from_angles(seq)
    final_error = (1 / suc * g_recon.IPoly - p).inf_norm
    print(f"Final error = {final_error}")
    if final_error < eps:
        return seq
    else:
        raise ValueError(
            "The angle finding program failed on given instance, with an error of {}. Please relax the error budget and/ or the success probability.".format(final_error))


def poly2laurent(pcoefs):
    """
    Convert polynomial coefficients to Laurent coefficients for polynomials
    with definit parity.

    Args:
        pcoefs: array of polynomial coefficients

    Returns:
        lcoefs: corresponding parity restricted Laurent coefficients

    Raises:
        AngleFindingError: if pcoefs does not have definite parity.
    """
    # convert polynomial coefficients to laurent
    ccoefs = np.polynomial.chebyshev.poly2cheb(pcoefs)

    # determine parity of polynomial
    is_even = np.max(np.abs(ccoefs[0::2])) > 1e-8
    is_odd = np.max(np.abs(ccoefs[1::2])) > 1e-8

    if is_even and is_odd:
        raise AngleFindingError(
            "Polynomial must have definite parity: {}".format(str(pcoefs)))

    if is_odd:
        lcoefs = ccoefs[1::2] / 2
        lcoefs = np.r_[lcoefs[::-1], lcoefs]
    else:
        lcoefs = ccoefs[0::2] / 2
        lcoefs = np.r_[lcoefs[-1:0:-1], 2 * lcoefs[0], lcoefs[1:]]

    return lcoefs


def QuantumSignalProcessingPhases(
        poly,
        eps=1e-4,
        suc=1 - 1e-4,
        signal_operator="Wx",
        measurement=None,
        tolerance=1e-6,
        method="laurent",
        **kwargs):
    """
    Generate QSP phase angles for a given polynomial.

    Args:
        poly: polynomial object
        eps: capitilization parameter for numerical stability
        suc: scaling factor for numerical stability
        signal_operator: QSP signal-dependent operation ['Wx', 'Wz']
        measurement: measurement basis (defaults to signal operator basis)
        tolerance: error tolerance in final reconstruction
        method: method to use for computing phase angles ['laurent', 'tf']

    Returns:
        Array of QSP angles.

    Raises:
        CompletionError: Raised when polynomial cannot be completed in given
            model.
        AngleFindingError: Raised if angle finding algorithm cannot find
        sequence to specified tolerance.
        ValueError: Raised if invalid model (or method) is specified.
    """
    if isinstance(poly, np.ndarray) or isinstance(poly, list):
        poly = Polynomial(poly)
    elif isinstance(poly, TargetPolynomial):
        poly = Polynomial(poly.coef)

    if measurement is None:
        if signal_operator == "Wx":
            measurement = "x"
        elif signal_operator == "Wz":
            measurement = "z"

    if method == "tf":
        if not signal_operator == "Wx":
            raise ValueError(
                f"Must use Wx signal operator model with tf method")
        return QuantumSignalProcessingPhasesWithTensorflow(poly,
                                                           measurement=measurement,
                                                           **kwargs)
    elif not method == "laurent":
        raise ValueError(f"Invalid method {method}")

    model = (signal_operator, measurement)

    # Perform completion
    if model in {("Wx", "x"), ("Wz", "z")}:
        # Capitalization: eps/2 amount of error budget is put to the highest
        # power for sake of numerical stability.
        poly = suc * \
            (poly + Polynomial([0, ] * poly.degree() + [eps / 2, ]))

        lcoefs = poly2laurent(poly.coef)
        lalg = completion_from_root_finding(lcoefs, coef_type="F")
    elif model == ("Wx", "z"):
        lalg = completion_from_root_finding(poly.coef, coef_type="P")
    else:
        raise ValueError(
            "Invalid model: {}".format(str(model))
        )

    # Decomposition phase
    phiset = angseq(lalg)

    # Verify by reconstruction
    adat = np.linspace(-1., 1., 100)
    response = ComputeQSPResponse(
        adat,
        phiset,
        signal_operator=signal_operator,
        measurement=measurement)["pdat"]
    expected = poly(adat)

    max_err = np.max(np.abs(response - expected))
    if max_err > tolerance:
        raise AngleFindingError(
            "The angle finding program failed on given instance, with an error of {}. Please relax the error budget and / or the success probability.".format(max_err))

    return phiset


def QuantumSignalProcessingPhasesWithTensorflow(
        poly,
        npts_theta=30,
        nepochs=5000,
        verbose=0,
        measurement="z",
        return_all=False):
    '''
    Compute QSP phase angles using optimization, with tensorflow, via the qsp_model submodule
    Running this imports pyqsp.qsp_models
    This import is done in the procedure because qsp_models requires tensorflow, and is heavyweight.
    We want to avoid requiring tensorflow to run pyqsp, if this is not needed.

    Args:
        poly: polynomial object, or StringPolynomial instance
        npts_theta: number of points to discretize theta axis into
        nepochs: number of epochs of training to do
        verbose: flag to give verbose debugging output during training
        return_all: if True, return dict with model, training history, and more

    Returns:
        Array of QSP angles (or if return_all: dict with model, training history, ...)
    '''
    import pyqsp.qsp_models as qsp_models
    import tensorflow as tf

    if not (
        isinstance(
            poly,
            Polynomial) or isinstance(
            poly,
            StringPolynomial)):
        raise ValueError(
            f"poly={poly} should be a Polynomial or StringPolynomial")

    poly_deg = poly.degree()

    # The intput theta training values
    th_in = np.arange(0, np.pi, np.pi / npts_theta)
    th_in = tf.reshape(th_in, (th_in.shape[0], 1))

    # The desired real part of p(x) which is the upper left value in the unitary of the qsp sequence
    # and the desired real part of q(x) = 0
    expected_outputs = [poly(np.cos(th_in)), np.zeros(th_in.shape[0])]

    # the tensorflow keras model
    model = qsp_models.construct_qsp_model(poly_deg, measurement=measurement)
    history = model.fit(
        x=th_in,
        y=expected_outputs,
        epochs=nepochs,
        verbose=verbose)
    phis = model.trainable_weights[0].numpy()
    if return_all:
        data = {'model': model,
                'history': history,
                'phis': phis,
                'th_in': th_in,
                'expected_outputs': expected_outputs,
                'poly': poly,
                }
        return data

    return phis



import os
import unittest

from pyqsp import main

# -----------------------------------------------------------------------------
# unit tests

test_cmds = [
    "--return-angles --poly=-1,0,2 poly2angles",
    "--return-angles --poly=-1,0,2 --plot --hide-plot poly2angles",
    "--plot-positive-only --plot-probability --plot-tight-y --plot-npts=400 --seqargs=10,0.5 fpsearch",
    "--plot-real-only --plot-npts=400 --seqargs=19,10 poly_sign",
    "--plot-real-only --plot-npts=400 --seqargs=3,0.3 invert",
    "--plot-real-only --plot-npts=400 --seqargs=10,0.1 hamsim",
    "--plot-real-only --plot-npts=400 --seqargs=18,10 poly_thresh",
    "--plot-real-only --plot-npts=400 --seqargs=18,10 poly_phase",
    "--plot-positive-only --plot-real-only --seqargs 30,0.3 efilter",
    "--plot-positive-only --plot-real-only --seqargs=20,3.5 gibbs",
    "--plot-real-only --seqargs=20,0.6,15 relu",
]


test_cmds_tf = [
    '--return-angles --func np.cos(3*x) --polydeg 6 polyfunc',
    '--plot-positive-only --plot --polyargs=19,10 --plot-real-only --polyname poly_sign --method tf poly',
]


class Test_main(unittest.TestCase):
    def test_main(self):
        for i, cmd in enumerate(test_cmds):
            with self.subTest(i=i):
                print(f"[pyqsp.test_main testing '{cmd}'")
                phiset = main.CommandLine(arglist=cmd.split(" "))

    def test_main_tf(self):
        '''
        These are slow tests, and run only if PYQSP_TEST_QSP_MODELS is set in the environment
        '''
        enabled = 'PYQSP_TEST_QSP_MODELS' in os.environ
        if not enabled:
            print(
                "[pyqsp.test] Skipping qsp_model tests: export PYQSP_TEST_QSP_MODELS=1 to enable these tests")
            return

        for i, cmd in enumerate(test_cmds_tf):
            with self.subTest(i=i):
                print(f"[pyqsp.test_main testing '{cmd}'")
                phiset = main.CommandLine(arglist=cmd.split(" "))



import numbers

import cirq
import numpy as np
import sympy
import tensorflow as tf

import tensorflow_quantum as tfq
from tensorflow_quantum.core.ops import tfq_unitary_op
from tensorflow_quantum.python import util
from tensorflow_quantum.python.layers.circuit_construction import elementary
from tensorflow_quantum.python.layers.circuit_executors import (
    expectation, input_checks, sampled_expectation)


class HybridControlledPQC(tf.keras.layers.Layer):
    """Hybrid Controlled Parametrized Quantum Circuit (PQC) Layer."""

    def __init__(self,
                 model_circuit,
                 operators,
                 *,
                 controlled_symbol_names=None,
                 native_symbol_names=None,
                 repetitions=None,
                 backend=None,
                 initializer=tf.keras.initializers.RandomUniform(0, 2 * np.pi),
                 regularizer=None,
                 constraint=None,
                 differentiator=None,
                 **kwargs):
        """Instantiate this layer.
        Create a layer that will output expectation values of the given
        operators when fed quantum data to it's input layer. This layer will
        take two input tensors, one representing a quantum data source (these
        circuits must not contain any symbols) and the other representing
        control parameters for the model circuit that gets appended to the
        datapoints.
        model_circuit: `cirq.Circuit` containing `sympy.Symbols` that will be
            used as the model which will be fed quantum data inputs.
        operators: `cirq.PauliSum` or Python `list` of `cirq.PauliSum` objects
            used as observables at the end of the model circuit.
        repetitions: Optional Python `int` indicating how many samples to use
            when estimating expectation values. If `None` analytic expectation
            calculation is used.
        backend: Optional Backend to use to simulate states. Defaults to
            the native TensorFlow simulator (None), however users may also
            specify a preconfigured cirq simulation object to use instead.
            If a cirq object is given it must inherit `cirq.SimulatesFinalState`
            if `sampled_based` is True or it must inherit `cirq.Sampler` if
            `sample_based` is False.
        differentiator: Optional `tfq.differentiator` object to specify how
            gradients of `model_circuit` should be calculated.
        """
        super().__init__(**kwargs)
        # Ingest model_circuit.
        if not isinstance(model_circuit, cirq.Circuit):
            raise TypeError("model_circuit must be a cirq.Circuit object."
                            " Given: ".format(model_circuit))
        self._symbols_list = list(
            sorted(util.get_circuit_symbols(model_circuit)))

#         self._native_symbols_list = list(sorted(controlled_symbol_names))
#         self._controlled_symbols_list = list(sorted(native_symbol_names))
#         self._native_symbols = tf.constant([str(x) for x in self._native_symbols_list])
#         self._controlled_symbols = tf.constant([str(x) for x in self._controlled_symbols_list])

        self._symbols = tf.constant(
            [str(x) for x in native_symbol_names + controlled_symbol_names])

        self._circuit = util.convert_to_tensor([model_circuit])

        if len(self._symbols_list) == 0:
            raise ValueError("model_circuit has no sympy.Symbols. Please "
                             "provide a circuit that contains symbols so "
                             "that their values can be trained.")

        # Ingest operators.
        if isinstance(operators, (cirq.PauliString, cirq.PauliSum)):
            operators = [operators]

        if not isinstance(operators, (list, np.ndarray, tuple)):
            raise TypeError("operators must be a cirq.PauliSum or "
                            "cirq.PauliString, or a list, tuple, "
                            "or np.array containing them. "
                            "Got {}.".format(type(operators)))
        if not all([
                isinstance(op, (cirq.PauliString, cirq.PauliSum))
                for op in operators
        ]):
            raise TypeError("Each element in operators to measure "
                            "must be a cirq.PauliString"
                            " or cirq.PauliSum")

        self._operators = util.convert_to_tensor([operators])

        # Ingest and promote repetitions.
        self._analytic = False
        if repetitions is None:
            self._analytic = True

        if not self._analytic and not isinstance(
                repetitions, numbers.Integral):
            raise TypeError("repetitions must be a positive integer value."
                            " Given: ".format(repetitions))

        if not self._analytic and repetitions <= 0:
            raise ValueError("Repetitions must be greater than zero.")

        if not self._analytic:
            self._repetitions = tf.constant(
                [[repetitions for _ in range(len(operators))]],
                dtype=tf.dtypes.int32)

        if not isinstance(
                backend,
                cirq.Sampler) and repetitions is not None and backend is not None:
            raise TypeError("provided backend does not inherit cirq.Sampler "
                            "and repetitions!=None. Please provide a backend "
                            "that inherits cirq.Sampler or set "
                            "repetitions=None.")

        if not isinstance(backend, cirq.SimulatesFinalState
                          ) and repetitions is None and backend is not None:
            raise TypeError("provided backend does not inherit "
                            "cirq.SimulatesFinalState and repetitions=None. "
                            "Please provide a backend that inherits "
                            "cirq.SimulatesFinalState.")

        # Ingest backend and differentiator.
        if self._analytic:
            self._layer = expectation.Expectation(
                backend=backend, differentiator=differentiator)
        else:
            self._layer = sampled_expectation.SampledExpectation(
                backend=backend, differentiator=differentiator)

        self._append_layer = elementary.AddCircuit()

        # create weights for only native symbols

        if not all(
                name in self._symbols_list for name in controlled_symbol_names):
            raise ValueError(
                "model_circuit does not contain all controlled symbol names ")

        # Set additional parameter controls.
        self.initializer = tf.keras.initializers.get(initializer)
        self.regularizer = tf.keras.regularizers.get(regularizer)
        self.constraint = tf.keras.constraints.get(constraint)

        # Weight creation is not placed in a Build function because the number
        # of weights is independent of the input shape.
        self._native_symbol_values = self.add_weight(
            'parameters',
            shape=(
                len(native_symbol_names),
            ),
            initializer=self.initializer,
            regularizer=self.regularizer,
            constraint=self.constraint,
            dtype=tf.float32,
            trainable=True)

    @property
    def symbols(self):
        """The symbols that are managed by this layer (in-order).
        Note: `symbols[i]` indicates what symbol name the managed variables in
            this layer map to.
        """
        return [sympy.Symbol(x) for x in self._symbols_list]

    def symbol_values(self):
        """Returns a Python `dict` containing symbol name, value pairs.
        Returns:
            Python `dict` with `str` keys and `float` values representing
                the current symbol values.
        """
        return dict(zip(self.symbols, self.get_weights()[0]))

    def build(self, input_shape):
        """Keras build function."""
        super().build(input_shape)

    def call(self, controlled_symbol_values):
        """Keras call function."""
        circuit_batch_dim = tf.gather(tf.shape(controlled_symbol_values), 0)
        tiled_up_model = tf.tile(self._circuit, [circuit_batch_dim])
        tiled_up_operators = tf.tile(self._operators, [circuit_batch_dim, 1])
        tiled_up_native_symbol_values = tf.tile(
            [self._native_symbol_values], [circuit_batch_dim, 1])
        symbol_values = tf.concat(
            [tiled_up_native_symbol_values, controlled_symbol_values], 1)
        # tiled_up_parameters = tf.tile(symbol_values, [circuit_batch_dim, 1])

        if self._analytic:
            return self._layer(tiled_up_model,
                               symbol_names=self._symbols,
                               symbol_values=symbol_values,
                               operators=tiled_up_operators)
        else:
            tiled_up_repetitions = tf.tile(self._repetitions,
                                           [circuit_batch_dim, 1])
            return self._layer(tiled_up_model,
                               symbol_names=self._symbols,
                               symbol_values=symbol_values,
                               operators=tiled_up_operators,
                               repetitions=tiled_up_repetitions)


class Unitary(tf.keras.layers.Layer):

    def __init__(self, **kwargs):
        """Instantiate a Unitary Layer.
        Create a layer that will calculate circuit unitary matrices and output
        them into the TensorFlow graph given a correct set of inputs.
        """
        super().__init__(**kwargs)
        self.unitary_op = tfq_unitary_op.get_unitary_op()
        self._w = None

    @tf.function
    def call(self, inputs,
             *,
             symbol_names=None,
             symbol_values=None,
             initializer=tf.keras.initializers.RandomUniform(0, 2 * np.pi)):
        """Keras call function.
        Input options:
            `inputs`, `symbol_names`, `symbol_values`:
                see `input_checks.expand_circuits`
        Output shape:
            `tf.RaggedTensor` with shape:
                [batch size of symbol_values, <size of state>, <size of state>]
                    or
                [number of circuits, <size of state>, <size of state>]
        """

        values_empty = False
        if symbol_values is None:
            values_empty = True

        inputs, symbol_names, symbol_values = input_checks.expand_circuits(
            inputs, symbol_names, symbol_values)

        circuit_batch_dim = tf.gather(tf.shape(inputs), 0)

        if values_empty:
            # No symbol_values were provided. So we assume the user wants us
            # to create and manage variables for them. We will do so by
            # creating a weights variable and tiling it up to appropriate
            # size of [batch, num_symbols].

            if self._w is None:
                # don't re-add variable.
                self._w = self.add_weight(name='circuit_learnable_parameters',
                                          shape=symbol_names.shape,
                                          initializer=initializer)

            symbol_values = tf.tile(tf.expand_dims(self._w, axis=0),
                                    tf.stack([circuit_batch_dim, 1]))

        unitary = self.unitary_op(inputs, symbol_names, symbol_values)
        return unitary.to_tensor()


class QSP(tf.keras.layers.Layer):
    """ QSP for """

    def __init__(self, poly_deg=0, **kwargs):
        super().__init__(**kwargs)
        self.q = cirq.GridQubit(0, 0)
        self.poly_deg = poly_deg
        self.symbol_names = [sympy.Symbol(f'phi{k}') for k in range(
            poly_deg + 1)] + [sympy.Symbol(f'th')]

        initializer = tf.keras.initializers.RandomUniform(0, 2 * np.pi)

        self.phi = self.add_weight(name='circuit_learnable_parameters',
                                   shape=(poly_deg + 1,),
                                   initializer=initializer)

    @tf.function
    def call(self, theta_inp):
        """Keras call function.
        Input options:
            `inputs`, `symbol_names`, `symbol_values`:
                see `input_checks.expand_circuits`
        Output shape:
            `tf.RaggedTensor` with shape:
                [batch size of symbol_values, <size of state>, <size of state>]
                    or
                [number of circuits, <size of state>, <size of state>]
        """

        wx = cirq.Circuit(cirq.rx(2 * theta_inp))
        self.rot_zs = [cirq.Circuit(cirq.rz(2 * self.phi[k])(self.q))
                       for k in range(self.poly_deg)]

        full_circuit = self.rot_zs[0]
        full_circuit_test = cirq.Circuit(
            cirq.rz(2 * self.symbol_names[0])(self.q),
            cirq.rx(2 * self.symbol_names[-1])(self.q),
            cirq.rz(2 * self.symbol_names[0])(self.q)
        )

        phi_values = tf.expand_dims(self.phi, axis=0)
        symbol_values = tf.expand_dims(tf.concat([self.phi, [4]], 0), 0)

        tensor_full_circuit_test = tfq.convert_to_tensor([full_circuit_test])
        tfq.from_tensor(tfq.resolve_parameters(
            tensor_full_circuit_test, self.symbol_names, symbol_values))

        return full_circuit_test.unitary()[0, 0]



import unittest

import numpy as np
from numpy.polynomial.polynomial import Polynomial

from pyqsp.angle_sequence import (AngleFindingError,
                                  QuantumSignalProcessingPhases, poly2laurent)
from pyqsp.LPoly import LPoly, PolynomialToLaurentForm

# -----------------------------------------------------------------------------
# unit tests


class Test_angle_sequence(unittest.TestCase):
    def test_poly2laurent_1(self):
        pcoefs = np.array([-3 - 2j, 0., 26 + 10j, 0., -24 - 8j])
        expected = np.array([-3 - 1j, 1 + 1j, 2., 1 + 1j, -3 - 1j]) / 2.
        result = poly2laurent(pcoefs)
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)

    def test_poly2laurent_2(self):
        pcoefs = np.array([0., -5 + 5j, 0., 8 - 4j])
        expected = np.array([2 - 1j, 1 + 2j, 1 + 2j, 2 - 1j]) / 2.
        result = poly2laurent(pcoefs)
        self.assertAlmostEqual(np.max(np.abs(expected - result)), 0.)

    def test_poly2laurent_3(self):
        pcoefs = np.array([1., -5 + 5j, 0., 8 - 4j])
        with self.assertRaises(AngleFindingError):
            poly2laurent(pcoefs)

    def test_response_1(self):
        pcoefs = [0, 1]

        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(poly, signal_operator="Wx")
        QuantumSignalProcessingPhases(poly, signal_operator="Wz")

    def test_response_2(self):
        pcoefs = [-1, 0, 2]
        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(poly, signal_operator="Wx")
        QuantumSignalProcessingPhases(poly, signal_operator="Wz")

    def test_response_3(self):
        pcoefs = [0, -3, 0, 4]
        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(poly, signal_operator="Wx")
        QuantumSignalProcessingPhases(poly, signal_operator="Wz")

    def test_response_4(self):
        pcoefs = [0., -2 + 1j, 0., 2.]
        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(
            poly, signal_operator="Wx", measurement="z")

    def test_response_5(self):
        pcoefs = [-1., 0., 50., 0., -400., 0., 1120., 0., -1280., 0., 512.]
        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(poly, signal_operator="Wx")
        QuantumSignalProcessingPhases(poly, signal_operator="Wz")
        QuantumSignalProcessingPhases(
            poly, signal_operator="Wx", measurement="z")

    def test_response_6(self):
        pcoefs = [-1., 0., (1 / 2) * (4 + 3j - (1 - 2j) * np.sqrt(3)),
                  0., (1 - 1j) * (-1j + np.sqrt(3))]
        poly = Polynomial(pcoefs)

        QuantumSignalProcessingPhases(
            poly, signal_operator="Wx", measurement="z")



import cirq
import numpy as np
import sympy
from cirq.contrib.svg import SVGCircuit


class QSPCircuit(cirq.Circuit):
    """QSP circuit

    A `cirq.Circuit` that implements the QSP sequence given by `phis`

    A tool to evaluate and visualize the response of a given QSP sequence.

    Allows substitution of arbitrary theta into the sequence
    """

    def __init__(self, phis):
        super(QSPCircuit, self).__init__()
        # recall that in the QSP sequence we rotate as exp(i * phi * Z), but
        # rz(theta) := exp(i * theta/2 * Z)
        self.phis = np.array(phis).flatten() * (-2)
        self.theta = sympy.Symbol("theta")
        self.q = cirq.GridQubit(0, 0)
        self._build_qsp_sequence(self.q)

    def _build_qsp_sequence(self, q):
        self.append(cirq.Circuit(cirq.rz(self.phis[0])(q)))
        for phi in self.phis[1:]:
            c = cirq.Circuit(cirq.rx(self.theta)(q), cirq.rz(phi)(q))
            self.append(c)

    def svg(self):
        """Get the SVG circuit (for visualization)"""
        return SVGCircuit(self)

    def qsp_response(self, thetas):
        """Evaluate the QSP response for a list of thetas

        params
        -----
        thetas: list of floats
            list of theta input of a QSP sequence

        returns
        -------
        numpy array with shape (len(params),)
            evaluates the qsp response Re[P(x)] + i * Re[Q(x)] * sqrt(1-x^2) from post selecting on |+> for each theta in thetas
        """
        return np.real(self.eval_px(thetas)) + \
            1j * np.real(self.eval_qx(thetas)) * np.sin(thetas)

    def eval_px(self, thetas):
        """Evaluate P(x) for a list of thetas

        params
        -----
        thetas: list of floats
            list of theta input of a QSP sequence

        returns
        -------
        numpy array with shape (len(params),)
            evaluates P(x) from the resulting QSP sequence for each theta in thetas
        """
        pxs = []
        for theta in np.array(thetas).flatten():
            resolver = cirq.ParamResolver({"theta": theta * (-2)})
            u = cirq.resolve_parameters(self, resolver).unitary()
            pxs.append(u[0, 0])
        return np.array(pxs)

    def eval_real_px(self, thetas):
        """Evaluate the QSP response (real part) for a list of thetas"""
        return np.real(self.eval_px(thetas))

    def eval_imag_px(self, thetas):
        """Evaluate the QSP response (imaginary part) for a list of thetas"""
        return np.imag(self.eval_px(thetas))

    def eval_qx(self, thetas):
        """Evaluate Q(x) for a list of thetas

        params
        -----
        thetas: list of floats
            list of theta input of a QSP sequence

        returns
        -------
        numpy array with shape (len(params),)
            evaluates Q(x) from the resulting QSP sequence for each theta in thetas
        """
        qxs = []
        for theta in np.array(thetas).flatten():
            resolver = cirq.ParamResolver({"theta": theta * (-2)})
            u = cirq.resolve_parameters(self, resolver).unitary()
            denom = np.sin(theta)
            if denom==0:
                denom = 1.0e-8
            qxs.append(u[0, 1] / (1j * denom))
        return np.array(qxs)



import unittest

import numpy as np

import pyqsp
from pyqsp import LPoly, angle_sequence, poly, response

# -----------------------------------------------------------------------------
# unit tests


class Test_poly(unittest.TestCase):

    def test_oneoverx0(self):
        pg = pyqsp.poly.PolyOneOverX()
        gp = pg.generate(8, 0.01, return_coef=False, ensure_bounded=False)
        # print(f"gp={gp}")
        print(f"gp(0.5) = {gp(0.5)}")
        assert abs(gp(0.5) - 2) < 0.01
        assert abs(gp(0.3) - 1 / (0.3)) < 0.01

    def test_oneoverx1(self):
        '''
        unit test to ensure that the polynomial approximation to 1/x really is close to 1/x
        '''
        kappa = 3
        epsilon = 0.01
        pg = pyqsp.poly.PolyOneOverX()
        gpoly = pg.generate(
            kappa,
            epsilon,
            return_coef=False,
            ensure_bounded=False)
        xpos = np.linspace(1 / kappa, 1)
        xval = np.concatenate([-xpos, xpos])
        expected = 1 / xval
        polyval = gpoly(xval)
        diff = abs(polyval - expected).mean()
        # print(f"diff={diff}")
        assert diff < 0.1

    def test_poly_one_over_x_response1(self):
        pg = pyqsp.poly.PolyOneOverX()
        pcoefs = pg.generate(3, 0.3, return_coef=True, ensure_bounded=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, signal_operator="Wx")
        # print(f"QSP angles = {phiset}")
        response.PlotQSPResponse(
            phiset, signal_operator="Wx", pcoefs=pcoefs, show=False)
        assert True

    def test_poly_sign1(self):
        pg = pyqsp.poly.PolySign()
        pfunc = pg.generate(17, 10)
        assert (pfunc(-0.9) < -0.2)
        assert (pfunc(0.9) > -0.2)

    def test_poly_thresh1(self):
        pg = pyqsp.poly.PolyThreshold()
        pcoefs = pg.generate(18, 10)
        pfunc = np.polynomial.Polynomial(pcoefs)
        # print(f"sign poly at -0.9 = {pfunc(-0.9)}")
        # print(f"sign poly at 0 = {pfunc(0)}")
        assert (pfunc(-0.9) < 0.1)
        assert (pfunc(0) > 0.3)
        assert (pfunc(0.9) < 0.1)

    def test_poly_phase1(self):
        pg = pyqsp.poly.PolyPhaseEstimation()
        pcoefs, scale = pg.generate(
            18, 10, ensure_bounded=True, return_scale=True)
        pfunc = np.polynomial.Polynomial(pcoefs)
        self.assertLess(pfunc(-0.9) / scale, -0.9)
        self.assertLess(pfunc(0.9) / scale, -0.9)
        self.assertGreater(pfunc(0) / scale, 0.9)

    def test_poly_gibbs1(self):
        pg = pyqsp.poly.PolyGibbs()
        pfunc = pg.generate(30, 4.5)
        # print(f"gibbs poly at 0.9 = {pfunc(0.9)}")
        # print(f"gibbs poly at 0 = {pfunc(0)}")
        assert (pfunc(0.9) < 0.3)
        assert (pfunc(0) > 0.9)

    def test_poly_cosine(self):
        tau = 10
        epsilon = 0.1
        pg = pyqsp.poly.PolyCosineTX()
        pcoefs, scale = pg.generate(
            tau,
            epsilon,
            return_coef=True,
            ensure_bounded=True,
            return_scale=True)
        pfunc = np.polynomial.Polynomial(pcoefs)

        x = np.linspace(-1, 1)
        self.assertTrue(
            np.max(np.abs(pfunc(x) - scale * np.cos(tau * x))) < epsilon)

    def test_poly_sine(self):
        tau = 10
        epsilon = 0.1
        pg = pyqsp.poly.PolySineTX()
        pcoefs, scale = pg.generate(
            tau,
            epsilon,
            return_coef=True,
            ensure_bounded=True,
            return_scale=True)
        pfunc = np.polynomial.Polynomial(pcoefs)

        x = np.linspace(-1, 1)
        self.assertTrue(
            np.max(np.abs(pfunc(x) - scale * np.sin(tau * x))) < epsilon)



'''
pyqsp/phases.py

Known QSP phases for specific responses, in the Wx convention
'''

import numpy as np


class PhaseGenerator:
    '''
    Abstract base class for phase generators
    '''

    def __init__(self, verbose=True):
        self.verbose = verbose
        return

    def help(self):
        '''
        return help text
        '''
        return "help text about the expected sequence arguments"

    def generate(self):
        '''
        return list of floats specifying the QSP phase angles
        '''
        return [0, 0]


class FPSearch(PhaseGenerator):
    '''
    Return phases for fixed point quantum search, following https://arxiv.org/abs/1409.3305

    d = length of search sequence (length of phase vector returned = 2*d)
    delta:  error probability allowed = 1-delta^2, at the fixed point
    gamma: can be specified instead of delta (it is calculated if delta given)
    return_alpha: if True, then return alpha_k instead of phivec
    '''

    def help(self):
        txt = """Return phases for fixed point quantum search, following https://arxiv.org/abs/1409.3305
    Arguments are:
        d = length of search sequence (length of phase vector returned = 2*d)
        delta:  error probability allowed = 1-delta^2, at the fixed point
        gamma: can be specified instead of delta (it is calculated if delta given)"""

        return txt

    def generate(self, d, delta=None, gamma=None, return_alpha=False):
        d = int(d)
        L = 2 * d + 1
        print(
            f"[pyqsp.fixed_point_search] generating length {2*d} sequence in the Wx convention")
        kvec = np.arange(1, d + 1)
        if gamma is None:
            if delta is None:
                delta = 0.1
            # T_{1/L}(1/delta)
            gamma = 1 / np.cosh((1 / L) * np.arccosh(1 / delta))
        sg = np.sqrt(1 - gamma**2)
        if self.verbose:
            print("[phi_fp]: gamma=%s" % gamma)
        avec = 2 * np.arctan2(1, (np.tan(2 * np.pi * kvec / L) * sg))
        if return_alpha:
            return avec
        bvec = - avec[::-1]
        self.avec = avec
        self.bvec = bvec
        phivec = np.zeros(2 * d)
        for k in range(d):
            # reverse order & scale to match QSVT convention
            phivec[2 * k] = -avec[d - k - 1] / 2
            phivec[2 * k + 1] = bvec[d - k - 1] / 2

        return phivec


class erf_step(PhaseGenerator):
    def help(self):
        return """Step function polynomial using erf(), but only for specific pre-computed values.  Argument is n, where n may be 7 or 23"""

    def generate(self, n):
        # n=7 poly approximation to erf(1.5 x), defined using W(x) convention
        # made for re(P)
        phi_n7_erf = [1.58019, 0.00172821, 0.251897, -
                      0.834542, -0.834542, 0.251897, 0.00172821, 0.00939863]

        # n=23 poly approximation to erf(2 x)
        # made for re(P)
        phi_n23_erf = [1.5708, 2.87883E-8, 5.83909E-7, 1.84144E-6, 0.0000209995,
                       -0.0000120126, 0.000564903, -0.0022922, 0.0150024,
                       -0.064666, 0.263754, -0.926685, -0.926912, 0.263756,
                       -0.0645576, 0.014932, -0.00225885, 0.000553565,
                       -8.29284E-6, 0.0000200459, 2.09476E-6, 5.3026E-7,
                       4.38578E-8, 3.66401E-9]
        if n == 7:
            return phi_n7_erf
        elif n == 23:
            return phi_n23_erf
        raise Exception("[pyqsp.phases.erf_step] n must be 7 or 23")

# -----------------------------------------------------------------------------


phase_generators = {'fpsearch': FPSearch,
                    'erf_step': erf_step,
                    }



import numpy
from scipy.linalg import toeplitz
from pyqsp.LPoly import LPoly, LAlg, w


def linear_system(g, ldeg):
    """
    Let vec(l) = l.IPoly.coefs + l.XPoly.coefs, i.e. to regard a Low Algebra element l of degree ldeg as a real vector in R^{2n+2}.
    M is the matrix representation of g in the sense that M * vec(l) = vec(l * g) for all l with degree ldeg.
    The linear system m, s consist of two parts; the first part ensures that the degree of g * l does not exceed deg - ldeg;
    the second part ensures that l(Id) = Id, fixing the SU(2) freedom left from the previous set of constraints.
    """
    deg = g.degree
    aligned_icoefs = g.IPoly.aligned(-deg, deg)
    aligned_xcoefs = g.XPoly.aligned(-deg, deg)

    def vec_to_mat(vec):
        return toeplitz(numpy.hstack((vec, [0] * ldeg)), [0] * (ldeg + 1))
    M = numpy.vstack((numpy.hstack((vec_to_mat(aligned_icoefs),
                                    vec_to_mat(-aligned_xcoefs[::-1]))),
                      numpy.hstack((vec_to_mat(aligned_xcoefs),
                                    vec_to_mat(aligned_icoefs[::-1])))))

    a = numpy.ones(ldeg + 1, dtype=float)
    b = numpy.zeros(ldeg + 1, dtype=float)
    m = numpy.vstack((numpy.hstack((a, b)), numpy.hstack(
        (b, a)), M[:ldeg, :], M[deg + 1: deg + 2 * ldeg + 1, :], M[-ldeg:, :]))
    s = numpy.zeros(m.shape[0], dtype=float)
    s[0] = 1
    return m, s


def decompose(g, ldeg):
    """
    Let
    g = exp(iθ_0 * X) * w * exp(iθ_1 * X) * ... w * exp(iθ_deg * X)
    One wants to solve for
    l = exp(iθ_0 * X) * w * exp(iθ_1 * X) * ... w * exp(-i(Σ_{i=1}^{deg-ldeg-1}θ_j) * X),
    and
    r = ~l * g = (exp(iΣ_{j=1}^{deg-ldeg}θ_j * X) * w * exp(iθ_{deg-ldeg+1} * X) * ... w * exp(iθ_deg * X).

    The linear system (m, s) is such that

    deg(l * g) <= deg - ldeg
    l(Id) = Id

    One can show that such a system has a unique solution.
    """
    deg = g.degree
    m, s = linear_system(g, ldeg)
    lstsq = numpy.linalg.lstsq(m, s, rcond=-1)
    v = lstsq[0]

    li = v[:ldeg + 1]
    lx = v[-ldeg - 1:]
    l = LAlg(LPoly(li, -ldeg), LPoly(lx, -ldeg))
    r = LAlg.truncate(l * g, -g.degree + ldeg, g.degree - ldeg)

    return ~l, r


def angseq(g):
    '''
    Divide-and-conquer approach to solve for the whole angle sequence.
    '''
    deg = g.degree
    if deg == 1:
        return g.left_and_right_angles
    else:
        l, r = decompose(g, deg // 2)
        a = angseq(l)
        b = angseq(r)
        return a[:-1] + [a[-1] + b[0]] + b[1:]



# visualization tools
import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt
import scipy
import numpy as np
from . import QSPCircuit


def compute_qsp_response(
        model=None,
        phis=None,
        return_all=False,
        show_svg=False):
    """Compute the QSP response againts the desired function response.

    Params
    ------
    f : function float --> float
            the desired function to be implemented by the QSP sequence
    model : Keras `Model` with `QSP` layer
            model trained to approximate f
    phis: numpy array of qsp phase angles (may be supplied instead of model)
    """
    all_th = np.arange(0, np.pi, np.pi / 300)

    # construct circuit
    if phis is None:
        phis = model.trainable_weights[0].numpy()
    qsp_circuit = QSPCircuit(phis)
    if show_svg:
        qsp_circuit.svg()
    circuit_px = qsp_circuit.eval_px(all_th)
    circuit_qx = qsp_circuit.eval_qx(all_th)
    qsp_response = qsp_circuit.qsp_response(all_th)
    if return_all:
        return qsp_response, all_th, circuit_px, circuit_qx
    return qsp_response


def plot_qsp_response(f, model=None, phis=None, title="QSP Response"):
    """Plot the QSP response againts the desired function response.

    Params
    ------
    f : function float --> float
            the desired function to be implemented by the QSP sequence
    model : Keras `Model` with `QSP` layer
            model trained to approximate f
    phis: numpy array of qsp phase angles (may be supplied instead of model)
    title: plot title (str)
    """
    qsp_response, all_th, circuit_px, circuit_qx = compute_qsp_response(
        model=model, phis=phis, return_all=True)

    pdata = {
        "x": np.cos(all_th),
        "Imag[p(x)]": np.imag(circuit_px),
        "Real[p(x)]": np.real(circuit_px),
        "Real[q(x)]": np.real(circuit_qx),
        "QSP Response": np.real(qsp_response)}
    if f is not None:
        pdata["desired f"] = f(np.cos(all_th))
    df = pd.DataFrame(pdata)
    df = df.melt("x", var_name="src", value_name="value")
    sns.lineplot(x="x", y="value", hue="src", data=df).set_title(title)
    plt.show()


def plot_loss(history):
    """Plot the error of a trained QSP model.

    Params
    ------
    history : tensorflow `History` object
    """
    plt.plot(history.history['loss'])
    plt.title("Learning QSP Angles")
    plt.xlabel("Iterations")
    plt.ylabel("Error")
    plt.show()



import argparse
import json
import sys

import numpy as np

import pyqsp
from pyqsp import angle_sequence, response
from pyqsp.phases import phase_generators
from pyqsp.poly import (StringPolynomial, TargetPolynomial,
                        polynomial_generators)

# -----------------------------------------------------------------------------


class VAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        curval = getattr(args, self.dest, 0) or 0
        values = values.count('v') + 1
        setattr(args, self.dest, values + curval)

# -----------------------------------------------------------------------------


def CommandLine(args=None, arglist=None):
    '''
    Main command line.  Accepts args, to allow for simple unit testing.
    '''
    import pkg_resources  # part of setuptools

    version = pkg_resources.require("pyqsp")[0].version
    help_text = """usage: pyqsp [options] cmd

Version: {}
Commands:

    poly2angles - compute QSP phase angles for the specified polynomial (use --poly)
    hamsim      - compute QSP phase angles for Hamiltonian simulation using the Jacobi-Anger expansion of exp(-i tau sin(2 theta))
    invert      - compute QSP phase angles for matrix inversion, i.e. a polynomial approximation to 1/a, for given delta and epsilon parameter values
    angles      - generate QSP phase angles for the specified --seqname and --seqargs
    poly        - generate QSP phase angles for the specified --polyname and --polyargs, e.g. sign and threshold polynomials
    polyfunc    - generate QSP phase angles for the specified --func and --polydeg using tensorflow + keras optimization method (--tf)
    response    - generate QSP polynomial response functions for the QSP phase angles specified by --phiset

Examples:

    pyqsp --poly=-1,0,2 poly2angles
    pyqsp --poly=-1,0,2 --plot poly2angles
    pyqsp --signal_operator=Wz --poly=0,0,0,1 --plot  poly2angles
    pyqsp --plot --tau 10 hamsim
    pyqsp --plot --tolerance=0.01 --seqargs 3 invert
    pyqsp --plot-npts=4000 --plot-positive-only --plot-magnitude --plot --seqargs=1000,1.0e-20 --seqname fpsearch angles
    pyqsp --plot-npts=100 --plot-magnitude --plot --seqargs=23 --seqname erf_step angles
    pyqsp --plot-npts=100 --plot-positive-only --plot --seqargs=23 --seqname erf_step angles
    pyqsp --plot-real-only --plot --polyargs=20,20 --polyname poly_thresh poly
    pyqsp --plot-positive-only --plot --polyargs=19,10 --plot-real-only --polyname poly_sign poly
    pyqsp --plot-positive-only --plot-real-only --plot --polyargs 20,3.5 --polyname gibbs poly
    pyqsp --plot-positive-only --plot-real-only --plot --polyargs 20,0.2,0.9 --polyname efilter poly
    pyqsp --plot-positive-only --plot --polyargs=19,10 --plot-real-only --polyname poly_sign --method tf poly
    pyqsp --plot --func "np.cos(3*x)" --polydeg 6 polyfunc
    pyqsp --plot --func "np.cos(3*x)" --polydeg 6 --plot-qsp-model polyfunc
    pyqsp --plot-positive-only --plot-real-only --plot --polyargs 20,3.5 --polyname gibbs --plot-qsp-model poly
    pyqsp --polydeg 16 --measurement="z" --func="-1+np.sign(1/np.sqrt(2)-x)+ np.sign(1/np.sqrt(2)+x)" --plot polyfunc

""".format(version)

    parser = argparse.ArgumentParser(
        description=help_text,
        formatter_class=argparse.RawTextHelpFormatter)

    def float_list(value):
        try:
            if not ',' in value and value.startswith("[") and value.endswith("]"):
                fstrset = value[1:-1].split(" ")
                fstrset = [x for x in fstrset if x]
                flist = list(map(float, fstrset))
                return flist
            return list(map(float, value.split(",")))
        except Exception as err:
            print(
                f"[pyqsp.float_list] failed to parse float list, err={err} from {value}")
            raise

    parser.add_argument("cmd", help="command")
    parser.add_argument(
        '-v',
        "--verbose",
        nargs=0,
        help="increase output verbosity (add more -v to increase versbosity)",
        action=VAction,
        dest='verbose')
    parser.add_argument("-o", "--output", help="output filename", default="")
    parser.add_argument(
        "--signal_operator",
        help="QSP sequence signal_operator, either Wx (signal is X rotations) or Wz (signal is Z rotations)",
        type=str,
        default="Wx")
    parser.add_argument(
        "--plot",
        help="generate QSP response plot",
        action="store_true")
    parser.add_argument(
        "--hide-plot",
        help="do not show plot (but it may be saved to a file if --output is specified)",
        action="store_true")
    parser.add_argument(
        "--return-angles",
        help="return QSP phase angles to caller",
        action="store_true")
    parser.add_argument(
        "--poly",
        help="comma delimited list of floating-point coeficients for polynomial, as const, a, a^2, ...",
        action="store",
        type=float_list)
    parser.add_argument(
        "--func",
        help="for tf method, numpy expression specifying ideal function (of x) to be approximated by a polynomial, e.g. 'np.cos(3 * x)'",
        type=str)
    parser.add_argument(
        "--polydeg",
        help="for tf method, degree of polynomial to use in generating approximation of specified function (see --func)",
        type=int)
    parser.add_argument(
        "--tau",
        help="time value for Hamiltonian simulation (hamsim command)",
        type=float,
        default=100)
    parser.add_argument(
        "--epsilon",
        help="parameter for polynomial approximation to 1/a, giving bound on error",
        type=float,
        default=0.3)
    parser.add_argument(
        "--seqname",
        help="name of QSP phase angle sequence to generate using the 'angles' command, e.g. fpsearch",
        type=str,
        default=None)
    parser.add_argument(
        "--seqargs",
        help="arguments to the phase angles generated by seqname (e.g. length,delta,gamma for fpsearch)",
        action="store",
        type=float_list)
    parser.add_argument(
        "--polyname",
        help="name of polynomial generate using the 'poly' command, e.g. 'sign'",
        type=str,
        default=None)
    parser.add_argument(
        "--polyargs",
        help="arguments to the polynomial generated by poly (e.g. degree,kappa for 'sign')",
        action="store",
        type=float_list)
    parser.add_argument(
        "--plot-magnitude",
        help="when plotting only show magnitude, instead of separate imaginary and real components",
        action="store_true")
    parser.add_argument(
        "--plot-probability",
        help="when plotting only show squared magnitude, instead of separate imaginary and real components",
        action="store_true")
    parser.add_argument(
        "--plot-real-only",
        help="when plotting only real component, and not imaginary",
        action="store_true")
    parser.add_argument(
        "--title",
        help="plot title",
        type=str,
        default=None)
    parser.add_argument(
        "--measurement",
        help="measurement basis if using the polyfunc argument",
        type=str,
        default=None)
    parser.add_argument(
        "--output-json",
        help="output QSP phase angles in JSON format",
        action="store_true")
    parser.add_argument(
        "--plot-positive-only",
        help="when plotting only a-values (x-axis) from 0 to +1, instead of from -1 to +1 ",
        action="store_true")
    parser.add_argument(
        "--plot-tight-y",
        help="when plotting scale y-axis tightly to real part of data",
        action="store_true")
    parser.add_argument(
        "--plot-npts",
        help="number of points to use in plotting",
        type=int,
        default=400)
    parser.add_argument(
        "--tolerance",
        help="error tolerance for phase angle optimizer",
        type=float,
        default=0.1)
    parser.add_argument(
        "--method",
        help="method to use for qsp phase angle generation, either 'laurent' (default) or 'tf' (for tensorflow + keras)",
        type=str,
        default='laurent')
    parser.add_argument(
        "--plot-qsp-model",
        help="show qsp_model version of response plot instead of the default plot",
        action="store_true")
    parser.add_argument(
        "--phiset",
        help="comma delimited list of QSP phase angles, to be used in the 'response' command",
        action="store",
        type=float_list)
    parser.add_argument(
        "--nepochs",
        help="number of epochs to use in tensorflow optimization",
        type=int,
        default=5000)
    parser.add_argument(
        "--npts-theta",
        help="number of discretized values of theta to use in tensorflow optimization",
        type=int,
        default=30)

    if not args:
        args = parser.parse_args(arglist)

    phiset = None
    plot_args = dict(plot_magnitude=args.plot_magnitude,
                     plot_probability=args.plot_probability,
                     plot_positive_only=args.plot_positive_only,
                     plot_real_only=args.plot_real_only,
                     plot_tight_y=args.plot_tight_y,
                     npts=args.plot_npts,
                     show=(not args.hide_plot),
                     show_qsp_model_plot=args.plot_qsp_model,
                     )

    qspp_args = dict(signal_operator=args.signal_operator,
                     method=args.method,
                     tolerance=args.tolerance,
                     nepochs=args.nepochs,
                     npts_theta=args.npts_theta,
                     )

    if args.cmd == "poly2angles":
        coefs = args.poly
        if not coefs:
            print(
                f"[pyqsp.main] must specify polynomial coeffients using --poly, e.g. --poly -1,0,2")
            sys.exit(0)
        if isinstance(coefs, str):
            coefs = list(map(float, coefs.split(",")))
        print(f"[pyqsp] polynomial coefficients={coefs}")
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            coefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                pcoefs=coefs,
                signal_operator=args.signal_operator,
                **plot_args)

    elif args.cmd == "hamsim":
        pg = pyqsp.poly.PolyCosineTX()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            return_coef=True,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * np.cos(args.seqargs[0] * x),
                signal_operator="Wx",
                title="Hamiltonian Simultation (Cosine)",
                **plot_args)

        pg = pyqsp.poly.PolySineTX()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            return_coef=True,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * np.sin(args.seqargs[0] * x),
                signal_operator="Wx",
                title="Hamiltonian Simultation (Sine)",
                **plot_args)

    elif args.cmd == "fpsearch":
        pg = pyqsp.phases.FPSearch()
        phiset = pg.generate(*args.seqargs)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                signal_operator="Wx",
                measurement="z",
                title="Oblivious amplification",
                **plot_args)

    elif args.cmd == "invert":
        pg = pyqsp.poly.PolyOneOverX()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            return_coef=True,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * 1 / x,
                signal_operator="Wx",
                measurement="z",
                title="Inversion",
                **plot_args)

    elif args.cmd == "gibbs":
        pg = pyqsp.poly.PolyGibbs()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * np.exp(-args.seqargs[1] * x),
                signal_operator="Wx",
                title="Gibbs distribution",
                **plot_args)

    elif args.cmd == "efilter":
        pg = pyqsp.poly.PolyEigenstateFiltering()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            delta = args.seqargs[1] / 2.
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale *
                (np.sign(x + delta) - np.sign(x - delta)) / 2,
                signal_operator="Wx",
                title="Eigenstate filtering",
                **plot_args)

    elif args.cmd == "relu":
        pg = pyqsp.poly.PolySoftPlus()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * np.maximum(x - args.seqargs[1], 0.),
                signal_operator="Wx",
                title="ReLU Function",
                **plot_args)

    elif args.cmd == "poly_sign":
        pg = pyqsp.poly.PolySign()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * np.sign(x),
                signal_operator="Wx",
                title="Sign Function",
                **plot_args)

    elif args.cmd == "poly_thresh":
        pg = pyqsp.poly.PolyThreshold()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale *
                (np.sign(x + 0.5) - np.sign(x - 0.5)) / 2,
                signal_operator="Wx",
                title="Threshold Function",
                **plot_args)

    elif args.cmd == "poly_phase":
        pg = pyqsp.poly.PolyPhaseEstimation()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale *
                (-1 + np.sign(1/np.sqrt(2) - x) + np.sign(1/np.sqrt(2) + x)),
                signal_operator="Wx",
                title="Phase Estimation Polynomial",
                **plot_args)

    elif args.cmd == "poly_rect":
        pg = pyqsp.poly.PolyRect()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale *
                1 - (np.sign(x + 1 / args.kappa) -
                     np.sign(x - 1 / args.kappa)) / 2,
                signal_operator="Wx",
                title="Rect Function",
                **plot_args)

    elif args.cmd == "invert_rect":
        pg = pyqsp.poly.PolyOneOverXRect()
        pcoefs, scale = pg.generate(
            *args.seqargs,
            ensure_bounded=True,
            return_scale=True)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=lambda x: scale * 1 / x,
                signal_operator="Wx",
                title="Poly Rect * 1/x",
                **plot_args)

    elif args.cmd == "poly":
        if not args.polyname or args.polyname not in polynomial_generators:
            print(
                f'Known polynomial generators: {",".join(polynomial_generators.keys())}')
            return
        pg = polynomial_generators[args.polyname]()
        if not args.polyargs:
            print(pg.help())
            return
        pcoefs = pg.generate(*args.polyargs)
        print(f"[pyqsp] polynomial coefs = {pcoefs}")
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            pcoefs, **qspp_args)
        if args.plot:
            target = None
            if isinstance(pcoefs, TargetPolynomial):
                target = pcoefs.target
            response.PlotQSPResponse(
                phiset,
                pcoefs=pcoefs,
                target=target,
                signal_operator="Wx",
                **plot_args)

    elif args.cmd == "angles":
        if not args.seqname or args.seqname not in phase_generators:
            print(
                f'Known phase generators: {",".join(phase_generators.keys())}')
            return
        pg = phase_generators[args.seqname]()
        if not args.seqargs:
            print(pg.help())
            return
        phiset = pg.generate(*args.seqargs)
        print(f"[pysqp] phiset={phiset}")
        if args.plot:
            response.PlotQSPResponse(phiset, signal_operator="Wx", **plot_args)

    elif args.cmd == "polyfunc":
        if (not args.func) or (not args.polydeg):
            print(f"Must specify --func and --polydeg")
            return
        qspp_args['method'] = 'tf'
        poly = StringPolynomial(args.func, args.polydeg)
        phiset = angle_sequence.QuantumSignalProcessingPhases(
            poly,
            measurement=args.measurement,
            **qspp_args)
        if args.plot:
            response.PlotQSPResponse(
                phiset,
                target=poly,
                signal_operator="Wx",
                measurement=args.measurement,
                title=args.title,
                **plot_args)

    elif args.cmd == "response":
        if not args.phiset:
            print("Must specify --phiset")
        phiset = args.phiset
        response.PlotQSPResponse(
            phiset, signal_operator=args.signal_operator, **plot_args)

    else:
        print(f"[pyqsp.main] Unknown command {args.cmd}")
        print(help_text)

    if (phiset is not None):
        if args.return_angles:
            return phiset
        if args.output_json:
            print(
                f"QSP Phase angles (for signal_operator={args.signal_operator}) in JSON format:")
            if not isinstance(phiset, list):
                phiset = phiset.tolist()
            print(json.dumps(phiset))



import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg


class ResponseError(Exception):
    pass


def ComputeQSPResponse(
        adat,
        phiset,
        signal_operator="Wx",
        measurement=None):
    """
    Compute QSP response.

    Args:
        adat: array of inputs to the polynomial
        phiset: array of QSP phases
        signal_operator: QSP signal-dependent operation ['Wx', 'Wz']
        measurement: measurement basis (defaults to signal operator basis)

    Returns:
        Response object.
    """
    pdat = []

    if measurement is None:
        if signal_operator == "Wx":
            measurement = "x"
        elif signal_operator == "Wz":
            measurement = "z"

    # define model parameters
    model = (signal_operator, measurement)
    if signal_operator == "Wx":
        def sig_op(a): return np.array(
            [[a, 1j * np.sqrt(1 - a**2)],
             [1j * np.sqrt(1 - a**2), a]])

        def qsp_op(phi): return np.array(
            [[np.exp(1j * phi), 0.],
             [0., np.exp(-1j * phi)]])
    elif signal_operator == "Wz":
        H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)

        def sig_op(a): return H @ np.array(
            [[a, 1j * np.sqrt(1 - a**2)],
             [1j * np.sqrt(1 - a**2), a]]) @ H

        def qsp_op(phi): return H @ np.array(
            [[np.exp(1j * phi), 0.],
             [0., np.exp(-1j * phi)]]) @ H
    else:
        raise ResponseError(
            "Invalid signal_operator: {}".format(signal_operator)
        )

    if measurement == "x":
        p_state = np.array([[1.], [1.]]) / np.sqrt(2)
    elif measurement == "z":
        p_state = np.array([[1.], [0.]])
    else:
        raise ResponseError(
            "Invalid measurement: {}".format(measurement)
        )

    # Compute response
    pmats = []
    for phi in phiset:
        pmats.append(qsp_op(phi))

    for a in adat:
        W = sig_op(a)
        U = pmats[0]
        for pm in pmats[1:]:
            U = U @ W @ pm
        pdat.append((p_state.T @ U @ p_state)[0, 0])

    pdat = np.array(pdat, dtype=np.complex128)

    ret = {'adat': adat,
           'pdat': pdat,
           'model': model,
           'phiset': phiset,
           }
    return ret


def PlotQSPResponse(
        phiset,
        signal_operator="Wx",
        measurement=None,
        npts=400,
        pcoefs=None,
        target=None,
        show=True,
        title=None,
        plot_magnitude=False,
        plot_probability=False,
        plot_positive_only=False,
        plot_real_only=False,
        plot_tight_y=False,
        show_qsp_model_plot=False):
    """
    Plot QSP response.

    Args:
        phiset: array of QSP phases
        signal_operator: QSP signal-dependent operation ['Wx', 'Wz']
        measurement: measurement basis (defaults to signal operator basis)
        npts: number of points to plot
        pcoefs: coefficients for expected polynomial response; will be plotted,
        if provided
        target: reference function, if provided
        show: call show function
        title: plot title, if provided
        plot_magnitude: if True, show magnitude instead of real and imaginary
            parts
        plot_probability: if True, show squared magnitude instead of real and
            imaginary parts
        plot_positive_only: if True, then only show positive ordinate values
        plot_real_only: if True, show only real part
        plot_tight_y: if True, set y-axis scale to be from min to max of real
            part; else go from +1.5 max to -1.5 max
        show_qsp_model_plot: if True, use qsp_model.plot_qsp_response

    Returns:
        Response object.
    """

    if show_qsp_model_plot:
        import pyqsp.qsp_models as qsp_models
        return qsp_models.plot_qsp_response(
            target, model=None, phis=phiset, title=title)

    if plot_positive_only:
        adat = np.linspace(0., 1., npts)
    else:
        adat = np.linspace(-1., 1., npts)

    qspr = ComputeQSPResponse(adat,
                              phiset,
                              signal_operator=signal_operator,
                              measurement=measurement)
    pdat = qspr['pdat']
    plt.figure(figsize=[8, 5])

    if pcoefs is not None:
        poly = pcoefs
        if not isinstance(poly, np.polynomial.Polynomial):
            poly = np.polynomial.Polynomial(pcoefs)
        expected = poly(adat)
        plt.plot(adat, expected, 'k-', label="target polynomial",
                 linewidth=3, alpha=0.5)

    if target is not None:
        L = np.max(np.abs(adat))
        xref = np.linspace(-L, L, 401)
        plt.plot(xref, target(xref), 'k--', label="target function",
                 linewidth=3, alpha=0.5)

    if plot_magnitude:
        plt.plot(adat, abs(pdat), 'k', label="abs[Poly(a)]")
        ymax = np.max(np.abs(pdat))
        ymin = np.min(np.abs(pdat))
    elif plot_probability:
        plt.plot(adat, abs(pdat)**2, 'k', label="abs[Poly(a)]^2")
        ymax = np.max(np.abs(pdat))**2
        ymin = np.min(np.abs(pdat))**2
    else:
        plt.plot(adat, np.real(pdat), 'k', label="Re[Poly(a)]")
        if not plot_real_only:
            plt.plot(adat, np.imag(pdat), 'b', label="Im[Poly(a)]")
        ymax = np.max(np.real(pdat))
        ymin = np.min(np.real(pdat))

    # format plot
    plt.ylabel("response")
    plt.xlabel("a")
    plt.legend(loc="upper right")

    if title is not None:
        plt.title(title)

    plt.xlim([np.min(adat), np.max(adat)])
    if plot_tight_y:
        plt.ylim([-0.1, 1.1 * ymax])
    else:
        plt.ylim([-1.5 * ymax, 1.5 * ymax])

    if show:
        plt.show()


def PlotQSPPhases(phiset, show=True):
    """
    Generate plot of QSP response function polynomial, i.e. Re( <0| U |0> )
    For values of model, see ComputeQSPResponse.

    pcoefs - coefficients for expected polynomial response; will be plotted,
        if provided
    target - reference function, if provided
    """
    plt.figure(figsize=[8, 5])

    plt.stem(phiset, markerfmt='bo', basefmt='k-')
    plt.xlabel("k")
    plt.ylabel("phi_k")
    plt.ylim([-np.pi, np.pi])

    if show:
        plt.show()



import numpy as np
import tensorflow as tf
from tensorflow import keras


class QSP(keras.layers.Layer):
    """Parameterized quantum signal processing layer.

    The `QSP` layer implements the quantum signal processing circuit with trainable QSP angles.
    The input of the layer is/are theta(s) where x = cos(theta), and w(x) is X rotation in the QSP sequence.

    The output is the real part of the upper left element in the resulting unitary that describes the whole sequence.
    This is Re[P(x)] in the representation of the QSP unitary from Gilyen et al.

    Input is of the form:
    [[theta1], [theta2], ... ]

    Output is of the form:
    [[P(x1)], [P(x2)], ...]

    The layer requires the desired polynomial degree of P(x)

    """

    def __init__(self, poly_deg=0, measurement="z"):
        """
        Params
        ------
        poly_deg: The desired degree of the polynomial in the QSP sequence.
            the layer will be parameterized with poly_deg + 1 trainable phi.
        measurement :
            measurement basis using the Wx model, {"x", "z"}
        """
        super(QSP, self).__init__()
        self.poly_deg = poly_deg
        phi_init = tf.random_uniform_initializer(minval=0, maxval=np.pi)
        self.phis = tf.Variable(
            initial_value=phi_init(shape=(poly_deg + 1, 1), dtype=tf.float32),
            trainable=True,
        )
        self.measurement = measurement

    def call(self, th):
        batch_dim = tf.gather(tf.shape(th), 0)

        # tiled up X rotations (input W(x))
        px = tf.constant([[0.0, 1], [1, 0]], dtype=tf.complex64)
        px = tf.expand_dims(px, axis=0)
        px = tf.repeat(px, [batch_dim], axis=0)

        rot_x_arg = tf.complex(real=0.0, imag=th)
        rot_x_arg = tf.expand_dims(rot_x_arg, axis=1)
        rot_x_arg = tf.tile(rot_x_arg, [1, 2, 2])

        wx = tf.linalg.expm(tf.multiply(px, rot_x_arg))

        # tiled up Z rotations
        pz = tf.constant([[1.0, 0], [0, -1]], dtype=tf.complex64)
        pz = tf.expand_dims(pz, axis=0)
        pz = tf.repeat(pz, [batch_dim], axis=0)

        z_rotations = []
        for k in range(self.poly_deg + 1):
            phi = self.phis[k]
            rot_z_arg = tf.complex(real=0.0, imag=phi)
            rot_z_arg = tf.expand_dims(rot_z_arg, axis=0)
            rot_z_arg = tf.expand_dims(rot_z_arg, axis=0)
            rot_z_arg = tf.tile(rot_z_arg, [batch_dim, 2, 2])

            rz = tf.linalg.expm(tf.multiply(pz, rot_z_arg))
            z_rotations.append(rz)

        u = z_rotations[0]
        for rz in z_rotations[1:]:
            u = tf.matmul(u, wx)
            u = tf.matmul(u, rz)

        # assume we are interested in the real part of p(x) and the real part of q(x) in
        # the resulting qsp unitary
        if self.measurement == "z":
            return tf.math.real(u[:, 0, 0]), tf.math.imag(u[:, 0, 0])
        elif self.measurement == "x":
            return tf.math.real(u[:, 0, 0]), tf.math.imag(u[:, 0, 1])
        else:
            raise ValueError(
                "Invalid measurement basis: {}".format(self.measurement))


def construct_qsp_model(poly_deg, measurement="z"):
    """Helper function that compiles a QSP model with mean squared error and adam optimizer.

    Params
    ------
    poly_deg : int
        the desired degree of the polynomial in the QSP sequence.
    measurement :
        measurement basis using the Wx model, {"x", "z"}

    Returns
    -------
    Keras model
        a compiled keras model with trainable phis in a poly_deg QSP sequence.
    """
    theta_input = tf.keras.Input(shape=(1,), dtype=tf.float32, name="theta")
    qsp = QSP(poly_deg, measurement=measurement)
    real_parts = qsp(theta_input)
    model = tf.keras.Model(inputs=theta_input, outputs=real_parts)
    optimizer = tf.keras.optimizers.Adam(learning_rate=0.1)
    loss = tf.keras.losses.MeanSquaredError()
    model.compile(optimizer=optimizer, loss=loss)
    return model



import os
import numpy as np
from pyqsp import LPoly

# -----------------------------------------------------------------------------
# unit tests

import unittest


class Test_lpoly(unittest.TestCase):

    def test_simple_unitary_from_angles1(self):
        phiset = [0]  # identity
        ualg = LPoly.LAlg.unitary_from_angles(phiset)
        print(f"For phiset={phiset}, U={ualg}")
        print(f"diagonal poly = {ualg.IPoly}")
        assert ualg.IPoly == LPoly.LPoly([1])

    def test_simple_unitary_from_angles2(self):
        phiset = [0, 0]  # w
        ualg = LPoly.LAlg.unitary_from_angles(phiset)
        print(f"For phiset={phiset}, U={ualg}")
        print(f"diagonal poly = {ualg.IPoly}")
        assert ualg.IPoly == LPoly.LPoly([0, 1], -1)

    def test_simple_unitary_from_angles3(self):
        phiset = [0, 0, 0]  # w^2
        ualg = LPoly.LAlg.unitary_from_angles(phiset)
        print(f"For phiset={phiset}, U={ualg}")
        print(f"diagonal poly = {ualg.IPoly}")
        assert ualg.IPoly == LPoly.LPoly([0, 0, 1], -2)

    def test_simple_unitary_from_angles4(self):
        # fourier transform of sine
        phiset = [-np.pi / 4, 0, np.pi / 4]
        ualg = LPoly.LAlg.unitary_from_angles(phiset)
        print(f"For phiset={phiset}, U={ualg}")
        print(f"diagonal poly = {ualg.IPoly}")
        assert ualg.IPoly == LPoly.LPoly([0.5, 0, 0.5], -2)

    def test_simple_unitary_from_angles5(self):
        # fourier transform of sine
        phiset = [-np.pi / 4, 0, 0, 0, np.pi / 4]
        ualg = LPoly.LAlg.unitary_from_angles(phiset)
        print(f"For phiset={phiset}, U={ualg}")
        print(f"diagonal poly = {ualg.IPoly}")
        assert ualg.IPoly == LPoly.LPoly([0.5, 0, 0, 0, 0.5], -4)

    def test_lpoly1(self):
        # Exp[-i (pi/2) X] * w = w^(-1) * iX
        w = LPoly.w
        Q0 = LPoly.LAlg.rotation(np.pi / 2)
        Q0.IPoly.round_zeros()
        prod = Q0 * w
        print(f"w = {w}")
        print(f"~w = {~w}")
        print(f"Q0 = {Q0}")
        print(f"Q0 * w = {prod}")
        print(f"poly([1,0], -1) = {LPoly.LPoly([1], -1)}")
        print(f"prod.XPoly coefs={prod.XPoly.coefs}, dmin={prod.XPoly.dmin}")
        assert prod.XPoly == LPoly.LPoly([1], -1)

    def test_LPoly_mul1(self):
        '''
        Test product of two LPoly's
        '''
        lp1 = LPoly.LPoly([1 / 2, 1 / 2], -1)
        lp2 = lp1 * lp1
        print(f"lp1={lp1}")
        print(f"lp1 * lp1 = {lp2}")
        print(f"lp2 coefs={lp2.coefs}")
        assert lp2.dmin == -2
        assert abs(lp2.coefs - np.array([0.25, 0.5, 0.25])).sum() < 1.0e-4


class Test_PolynomialToLaurentForm(unittest.TestCase):

    def test_p2lf3(self):
        pcoefs = [-1, 0, 2]
        Plp = LPoly.PolynomialToLaurentForm(pcoefs)
        print(f"Polynomial {pcoefs} -> Laurent poly {Plp}")
        print(f"LPoly coefs = {Plp.coefs}")
        assert Plp.dmin == -2
        assert abs(Plp.coefs - np.array([0.5, 0, 0.5])).sum() < 1.0e-4

    def test_p2lf1(self):
        coefs = [0, 1]
        lp = LPoly.PolynomialToLaurentForm(coefs)
        print(f"Polynomial {coefs} -> Laurent poly {lp}")
        assert lp.dmin == -1
        assert abs(lp.coefs - np.array([1 / 2, 1 / 2])).sum() < 1.0e-8

    def test_p2lf2(self):
        '''
        Note how LPoly coefficient list elements correspond to powers of w which go up by 2 at a time, not one!
        '''
        coefs = [0, -3, 0, 4]
        lp = LPoly.PolynomialToLaurentForm(coefs)
        print(f"Polynomial {coefs} -> Laurent poly {lp}")
        print(f"LPoly coefs = {lp.coefs}")
        assert lp.dmin == -3
        assert abs(lp.coefs - np.array([0.5, 0, 0, 0.5])).sum() < 1.0e-8



import numpy
PRES = 5000


class LPoly():
    '''
    Laurent polynomials with parity constraint.
    '''

    def __init__(self, coefs, dmin=0):
        '''
        coefs = one-dimensional list or array of coefficients for the Laurent polynomial
        dmin = minimum power for the polynomial, e.g. 0 or -1 or -4
        '''
        self.coefs = numpy.array(coefs)
        if len(self.coefs) == 0:
            self.dmin = dmin
            self.iszero = True
            self.coefs = [0]
        else:
            assert(len(self.coefs.shape) == 1), self.coefs
            self.dmin = dmin
            self.iszero = False

    @property
    def dmax(self):
        return 2 * len(self.coefs) + self.dmin - 2

    @property
    def degree(self):
        """
        The degree of a Laurent polynomial is defined to be the maximum absolute value of the powers over all nonzero terms.
        """
        return max(-self.dmin, self.dmax)

    @property
    def norm(self):
        """
        The norm of a Laurent polynomial is defined to be the l2 norm of the coefficients regarded as a vector.
        """
        return numpy.linalg.norm(self.coefs)

    @property
    def inf_norm(self):
        """
        The infinity norm of a Laurent polynomial is defined to be the maximum modulus over the unit circle.
        """
        i, x = self.curve
        return numpy.amax(
            numpy.absolute(i + 1j * x))

    @property
    def parity(self):
        '''
        Parity of the polynomial. 0 for even parity Laurent polynomials and 1 for odd parity Laurent polynomials.
        '''
        return self.dmin % 2

    @property
    def curve(self):
        values = numpy.exp(
            numpy.outer(
                numpy.linspace(
                    -self.parity * 1j * numpy.pi,
                    1j * numpy.pi,
                    PRES),
                range(
                    self.dmin,
                    self.dmax + 1,
                    2))).dot(
            self.coefs)
        return numpy.real(values), numpy.imag(values)

    def __getitem__(self, key):
        if (key - self.dmin) % 2:
            return 0
        pos = (key - self.dmin) // 2
        if (pos < len(self.coefs)) and (pos >= 0):
            return self.coefs[pos]
        else:
            return 0

    def __mul__(self, other):
        if isinstance(other, LAlg):
            return LAlg(self * other.IPoly, self * other.XPoly)
        if not isinstance(other, LPoly):
            return LPoly(other * self.coefs, self.dmin)
        if self.iszero or other.iszero:
            return LPoly([])
        dmin = self.dmin + other.dmin
        coefs = numpy.convolve(self.coefs, other.coefs)
        return LPoly(coefs, dmin)

    def __rmul__(self, other):
        if isinstance(other, LAlg):
            return LAlg(self * other.IPoly, ~self * other.XPoly)
        elif not isinstance(other, LPoly):
            return LPoly(other * self.coefs, self.dmin)

    def __add__(self, other):
        '''
        Only Laurent polynomials of the same parity can be added together in order to preserve parity.
        '''
        if self.iszero:
            return LPoly(other.coefs, other.dmin)
        if other.iszero:
            return LPoly(self.coefs, self.dmin)
        assert (self.parity == other.parity), "not of the same parity"
        dmin = min(self.dmin, other.dmin)
        dmax = max(self.dmax, other.dmax)
        coefs = self.aligned(dmin, dmax) + other.aligned(dmin, dmax)
        return LPoly(coefs, dmin)

    def __neg__(self):
        return LPoly(-1 * self.coefs, self.dmin)

    def __invert__(self):
        '''
        Conjugation of a Laurent polynomial maps f(w) to f(w^-1).
        '''
        dmin = -self.dmax
        coefs = self.coefs[::-1]
        return LPoly(coefs, dmin)

    def __sub__(self, other):
        return self + (-other)

    def __str__(self):
        return " + ".join(["{} * w ^ ({})".format(self.coefs[i],
                                                  self.dmin + 2 * i)
                           for i in range(len(self.coefs))])

    def aligned(self, dmin, dmax):
        if(self.iszero):
            return numpy.zeros((dmax - dmin) // 2 + 1)
        else:
            assert (
                dmin <= self.dmin) and (
                dmax >= self.dmax), "interval not valid"
            return numpy.hstack((numpy.zeros((self.dmin - dmin) // 2),
                                 numpy.array(self.coefs),
                                 numpy.zeros((dmax - self.dmax) // 2)))

    def eval(self, angles):
        '''
        Evalute the Laurent polynomial f(w) at w = exp(i * angle) for angle iterating over angles. Returns a complex array .
        '''
        if self.iszero:
            return 1
        res = self.coefs.dot(
            numpy.exp(
                1j *
                numpy.outer(
                    numpy.arange(
                        self.dmin,
                        self.dmax +
                        1,
                        2),
                    angles)))
        return res

    @classmethod
    def truncate(cls, p, dmin, dmax):
        lb = min(dmin, p.dmin)
        ub = max(dmax, p.dmax)
        return LPoly(p.aligned(lb, ub + 2)
                     [(dmin - lb) // 2:(dmax - ub) // 2 - 1], dmin)

    @classmethod
    def isconsistent(cls, a, b):
        if a.iszero:
            return True
        if b.iszero:
            return True
        return a.parity == b.parity

    def __eq__(self, other):
        '''
        Equality test between two LPoly instances.
        Note that this doesn't check for when two LPoly's are equal but have different dmin
        and different coefficients, e.g. because of zeros in the coefficients
        '''
        iseq = abs(self.coefs - other.coefs).sum() < 0.001
        iseq = iseq and (self.dmin == other.dmin)
        return iseq

    def round_zeros(self, thresh=1.0e-5):
        '''
        round small coefficients down to zero
        '''
        self.coefs[self.coefs < thresh] = 0

    def pos_half(self):
        '''
        Return Laurent polynomial with coefficients of only positive exponents
        The negative (and zero) exponent coefficients are set to zero.
        '''
        new_coefs = numpy.copy(self.coefs)
        nhalf = int(numpy.ceil(len(new_coefs) / 2))
        new_coefs[:nhalf] = 0
        return LPoly(new_coefs, self.dmin)

    def neg_half(self):
        '''
        Return Laurent polynomial with coefficients of only negative (or zero) exponents
        The positive exponent coefficients are set to zero.
        '''
        new_coefs = numpy.copy(self.coefs)
        nhalf = int(numpy.ceil(len(new_coefs) / 2))
        new_coefs[nhalf:] = 0
        return LPoly(new_coefs, self.dmin)


class LAlg():
    '''
    Low algebra elements with parity constraints.
    A Low algebra element g is a matrix-valued Laurent polynomial with the given form
    g(w) = IPoly(w) + XPoly(w) * iX,
    where IPoly and XPoly are real Laurent polynomials sharing the same parity,
    and the elements w and X satisfy the relations (iX)^2 = Id and XwXw = Id.
    '''

    def __init__(self, IPoly=LPoly([], 0), XPoly=LPoly([], 0)):
        self.IPoly = IPoly
        self.XPoly = XPoly
        assert LPoly.isconsistent(self.IPoly, self.XPoly),\
            "The algebra element does not have a consistent parity"

    @property
    def degree(self):
        return max(self.IPoly.degree, self.XPoly.degree)

    @property
    def norm(self):
        return numpy.sqrt(self.IPoly.norm**2 + self.XPoly.norm**2)

    @property
    def parity(self):
        return self.IPoly.parity

    def __str__(self):
        res = []
        if not self.IPoly.iszero:
            res.append(str(self.IPoly))
        if not self.XPoly.iszero:
            res.append("( " + str(self.XPoly) + " ) * iX")
        return " + ".join(res)

    def __add__(self, other):
        if isinstance(other, LPoly):
            return LAlg(self.IPoly + other, self.XPoly)
        return LAlg(self.IPoly + other.IPoly, self.XPoly + other.XPoly)

    def __neg__(self):
        return LAlg(-self.IPoly, -self.XPoly)

    def __sub__(self, other):
        return self + (-other)

    def __invert__(self):
        '''
        Conjugation of g(w) = A(w) + B(w) * iX is ~g(w) = A(w^-1) - B(w) * iX.
        '''
        return LAlg(~self.IPoly, -self.XPoly)

    def __mul__(self, other):
        if isinstance(other, LPoly):
            return LAlg(self.IPoly * other, self.XPoly * ~other)
        if not isinstance(other, LAlg):
            return LAlg(self.IPoly * other, self.XPoly * other)
        return LAlg(self.IPoly * other.IPoly - self.XPoly * (~other.XPoly),
                    self.IPoly * other.XPoly + self.XPoly * (~other.IPoly))

    @property
    def pnorm(self):
        return (self * (~self)).IPoly

    @property
    def unitarity(self):
        '''
        Unitarity of an element A is defined to be ||Id-a*~a||_2^2.
        '''
        return (LPoly([1]) - self.pnorm).norm

    @property
    def angle(self):
        '''
        For degree 0 element M, the corresponding angle is defined such that M \propto exp(angle * iX).
        '''
        assert (self.degree == 0), "deg = {}".format(self.degree)
        return numpy.angle(self.IPoly[0] + self.XPoly[0] * 1j)

    @property
    def left_and_right_angles(self):
        '''
        For a degree 1 element g(w) = exp(a * iX) * w * exp(b * iX), return the left and right rotation angles [a, b].
        Note that g(Id) = exp((a + b) * iX) and g(iZ) = exp((a-b) * iX) * iZ.
        '''
        assert self.degree == 1
        summation = numpy.angle(
            self.IPoly.eval(0)[0] +
            1j *
            self.XPoly.eval(0)[0])
        difference = numpy.angle(self.IPoly.eval(
            numpy.pi / 2)[0] - 1j * self.XPoly.eval(numpy.pi / 2)[0]) - numpy.pi / 2
        res = [(summation + difference) / 2, (summation - difference) / 2]
        return res

    @classmethod
    def rotation(cls, ang):
        '''
        Degree 0 rotation with a certain angle. Inverse of `LAlg.angle`.
        '''
        return LAlg(LPoly([numpy.cos(ang)]), LPoly([numpy.sin(ang)]))

    @property
    def curve(self):
        i, z = self.IPoly.curve
        y, x = self.XPoly.curve
        return numpy.angle(1j * i - y) * numpy.sqrt(1 - x**2 - z**2), z, x

    @classmethod
    def truncate(cls, g, dmin, dmax):
        return LAlg(LPoly.truncate(g.IPoly, dmin, dmax),
                    LPoly.truncate(g.XPoly, dmin, dmax))

    @classmethod
    def generator(cls, ang):
        '''
        Generator elements of the unitary group; each one is w conjugated by some rotation.
        '''
        return cls.rotation(ang) * w * cls.rotation(-ang)

    @classmethod
    def unitary_from_conjugations(cls, ang):
        '''
        Generating a special unitary element from the conjugation angles.
        '''
        res = Id
        for i in ang:
            res *= cls.generator(i)
        return res

    @classmethod
    def unitary_from_angles(cls, ang):
        '''
        Generating a unitary element from the rotation angles sandwiching w.
        '''
        res = cls.rotation(ang[0])
        for i in ang[1:]:
            res = res * w * cls.rotation(i)
        return res

# -----------------------------------------------------------------------------


def PolynomialToLaurentForm(coefs):
    '''
    Given coefficients of a polynomial in a = cos(theta), return the Laurent form of the polynomial, as an instance of LPoly
    coefs[0] = constant term, coefs[1] ~= a, coefs[2] ~= a^2, ...
    We substitute a = (w + 1/w)/2, based on w = exp[i * theta], where theta = arccos[a]
    Note i sqrt(1-a^2) = (w - 1/w)/2
    '''
    lp = LPoly([])
    for k, c in enumerate(coefs):
        if c == 0:
            continue
        if k == 0:
            lpcoefs = [1]
            dmin = 0
        else:
            lpcoefs = [1 / 2, 1 / 2]
            dmin = -1
        nlp = LPoly(lpcoefs, dmin)
        for j in range(k - 1):
            nlp = nlp * LPoly(lpcoefs, dmin)
        lp = lp + c * nlp
    return lp

# -----------------------------------------------------------------------------
# Definition of elements


Id = LPoly([1])
w = LPoly([1], 1)
iX = LAlg(XPoly=LPoly([1]))



import numpy as np
import scipy.optimize
import scipy.special
from scipy.interpolate import approximate_taylor_polynomial

# -----------------------------------------------------------------------------


class StringPolynomial:
    '''
    Representation of a polynomial using a python string which specifies a numpy function,
    and an integer giving the desired polynomial's degree.
    '''

    def __init__(self, funcstr, poly_deg):
        '''
        funcstr: (str) specification of function using "x" as the argument, e.g. "np.where(x<0, -1 ,np.where(x>0,1,0))"
                 The function should accept a numpy array as "x"
        poly_deg: (int) degree of the polynoimal to be used to approximate the specified function
        '''
        self.funcstr = funcstr
        self.poly_deg = int(poly_deg)
        try:
            self.__call__(0.5)
        except Exception as err:
            raise ValueError(
                f"Invalid function specifciation, failed to evaluate at x=0.5, err={err}")

    def degree(self):
        return self.poly_deg

    def __call__(self, arg):
        ret = eval(self.funcstr, globals(), {'x': arg})
        return ret

    def target(self, arg):
        return self.__call__(arg)

# -----------------------------------------------------------------------------


class TargetPolynomial(np.polynomial.Polynomial):
    '''
    Polynomial with ideal target
    '''

    def __init__(self, *args, target=None, scale=None, **kwargs):
        '''
        target = function which accepts argument and gives ideal response, e.g. lambda x: x**2
        scale = metadata about scale of polynomial
        '''
        self.target = target
        self.scale = scale
        super().__init__(*args, **kwargs)

# -----------------------------------------------------------------------------


class PolyGenerator:
    '''
    Abstract base class for polynomial generators
    '''

    def __init__(self, verbose=True):
        self.verbose = verbose
        return

    def help(self):
        '''
        return help text
        '''
        return "help text about the expected polynomial arguments"

    def generate(self):
        '''
        return list of floats specifying the [const, a, a^2, ...] coefficients of the polynomial
        '''
        return [0, 0]

# -----------------------------------------------------------------------------


class PolyCosineTX(PolyGenerator):

    def help(self):
        return "Used for Hamiltonian simultion for time tau. Error is epsilon"

    def generate(
            self,
            tau=10.,
            epsilon=0.1,
            return_coef=True,
            ensure_bounded=True,
            return_scale=False):
        '''
        Approximation to cos(tx) polynomial, using sums of Chebyshev
        polynomials, from Optimal Hamiltonian Simulation by Quantum Signal
        Processing by Low and Chuang,
        https://arxiv.org/abs/1606.02685
        ensure_bounded: True if polynomial should be normalized to be between
        +/- 1
        '''
        r = scipy.optimize.fsolve(lambda r: (
            np.e * np.abs(tau) / (2 * r))**r - (5 / 4) * epsilon, tau)[0]
        print(r)
        R = np.floor(r / 2).astype(int)
        R = max(R, 1)

        print(f"R={R}")

        g = scipy.special.jv(0, tau) * np.polynomial.chebyshev.Chebyshev([1])
        for k in range(1, R + 1):
            gcoef = 2 * scipy.special.jv(2 * k, tau)
            deg = 2 * k
            g += (-1)**k * gcoef * \
                np.polynomial.chebyshev.Chebyshev([0] * deg + [1])

        if ensure_bounded:
            scale = 0.5
            g = scale * g
            print(f"[PolyCosineTX] rescaling by {scale}.")

        if return_coef:
            pcoefs = np.polynomial.chebyshev.cheb2poly(g.coef)
            if ensure_bounded and return_scale:
                return pcoefs, scale
            else:
                return pcoefs
        return g


class PolySineTX(PolyGenerator):

    def help(self):
        return "Used for Hamiltonian simultion for time tau. Error is epsilon"

    def generate(
            self,
            tau=10.,
            epsilon=0.1,
            return_coef=True,
            ensure_bounded=True,
            return_scale=False):
        '''
        Approximation to cos(tx) polynomial, using sums of Chebyshev
        polynomials, from Optimal Hamiltonian Simulation by Quantum Signal
        Processing by Low and Chuang,
        https://arxiv.org/abs/1606.02685
        ensure_bounded: True if polynomial should be normalized to be between
        +/- 1
        '''
        r = scipy.optimize.fsolve(lambda r: (
            np.e * np.abs(tau) / (2 * r))**r - (5 / 4) * epsilon, tau)[0]
        print(r)
        R = np.floor(r / 2).astype(int)
        R = max(R, 1)

        print(f"R={R}")

        g = np.polynomial.chebyshev.Chebyshev([0])
        for k in range(0, R + 1):
            gcoef = 2 * scipy.special.jv(2 * k + 1, tau)
            deg = 2 * k + 1
            g += (-1)**k * gcoef * \
                np.polynomial.chebyshev.Chebyshev([0] * deg + [1])

        if ensure_bounded:
            scale = 0.5
            g = scale * g
            print(f"[PolySineTX] rescaling by {scale}.")

        if return_coef:
            pcoefs = np.polynomial.chebyshev.cheb2poly(g.coef)
            if ensure_bounded and return_scale:
                return pcoefs, scale
            else:
                return pcoefs
        return g

# -----------------------------------------------------------------------------


class PolyOneOverX(PolyGenerator):

    def help(self):
        return "Region of validity is from 1/kappa to 1, and from -1/kappa to -1.  Error is epsilon"

    def generate(
            self,
            kappa=3,
            epsilon=0.1,
            return_coef=True,
            ensure_bounded=True,
            return_scale=False):
        '''
        Approximation to 1/x polynomial, using sums of Chebyshev polynomials,
        from Quantum algorithm for systems of linear equations with exponentially
        improved dependence on precision, by Childs, Kothari, and Somma,
        https://arxiv.org/abs/1511.02306v2

        Define region D_kappa to be from 1/kappa to 1, and from -1/kappa to -1.  A good
        approximation is desired only in this region.

        ensure_bounded: True if polynomial should be normalized to be between +/- 1
        '''
        b = int(kappa**2 * np.log(kappa / epsilon))
        j0 = int(np.sqrt(b * np.log(4 * b / epsilon)))
        print(f"b={b}, j0={j0}")

        g = np.polynomial.chebyshev.Chebyshev([0])
        for j in range(j0 + 1):
            gcoef = 0
            for i in range(j + 1, b + 1):
                gcoef += scipy.special.binom(2 * b, b + i) / 2**(2 * b)
            deg = 2 * j + 1
            g += (-1)**j * gcoef * \
                np.polynomial.chebyshev.Chebyshev([0] * deg + [1])
        g = 4 * g

        if ensure_bounded:
            res = scipy.optimize.minimize(g, (-0.1,), bounds=[(-0.8, 0.8)])
            pmin = res.x
            print(
                f"[PolyOneOverX] minimum {g(pmin)} is at {pmin}: normalizing")
            scale = 1 / abs(g(pmin))
            if 0:
                scale = scale * 0.9
            else:
                scale = scale * 0.5
                print("[PolyOneOverX] bounding to 0.5")
            g = scale * g

        if return_coef:
            if 1:
                pcoefs = np.polynomial.chebyshev.cheb2poly(g.coef)
            else:
                pcoefs = g.coef
            print(f"[pyqsp.PolyOneOverX] pcoefs={pcoefs}")
            if ensure_bounded and return_scale:
                return pcoefs, scale
            else:
                return pcoefs

        return g


class PolyOneOverXRect(PolyGenerator):

    def help(self):
        return "Region of validity is from 1/kappa to 1, and from -1/kappa to -1.  Error is epsilon"

    def generate(
            self,
            degree=6,
            delta=2,
            kappa=3,
            epsilon=0.1,
            ensure_bounded=True,
            return_scale=False):

        coefs_invert, scale1 = PolyOneOverX().generate(2 * kappa,
                                                       epsilon,
                                                       ensure_bounded,
                                                       return_scale=True)

        coefs_rect, scale2 = PolyRect().generate(degree,
                                                 delta,
                                                 kappa,
                                                 ensure_bounded,
                                                 return_scale=True)

        poly_invert = np.polynomial.Polynomial(coefs_invert)
        poly_rect = np.polynomial.Polynomial(coefs_rect)

        pcoefs = (poly_invert * poly_rect).coef

        if return_scale:
            return pcoefs, scale1 * scale2
        else:
            return pcoefs


# -----------------------------------------------------------------------------


class PolyTaylorSeries(PolyGenerator):
    '''
    Base class for PolySign and PolyThreshold
    '''

    def taylor_series(
            self,
            func,
            degree,
            ensure_bounded=True,
            return_scale=False,
            npts=100,
            max_scale=0.5):
        '''
        Return numpy Polynomial approximation for func, constructed using
        taylor series, of specified degree.
        Evaluate approximation using mean absolut difference on npts points in
        the domain from -1 to 1.
        '''
        the_poly = approximate_taylor_polynomial(func, 0, degree, 1)
        the_poly = np.polynomial.Polynomial(the_poly.coef[::-1])
        if ensure_bounded:
            res = scipy.optimize.minimize(-the_poly, (0.1,), bounds=[(-1, 1)])
            pmax = res.x
            scale = 1 / abs(the_poly(pmax))
            # use this for the new QuantumSignalProcessingWxPhases code, which
            # employs np.polynomial.chebyshev.poly2cheb(pcoefs)
            scale = scale * max_scale
            print(f"[PolyTaylorSeries] max {scale} is at {pmax}: normalizing")
            the_poly = scale * the_poly
        adat = np.linspace(-1, 1, npts)
        pdat = the_poly(adat)
        edat = func(adat)
        avg_err = abs(edat - pdat).mean()
        print(
            f"[PolyTaylorSeries] average error = {avg_err} in the domain [-1, 1] using degree {degree}")
        if ensure_bounded and return_scale:
            return the_poly, scale
        else:
            return the_poly

# -----------------------------------------------------------------------------


class PolySign(PolyTaylorSeries):

    def help(self):
        return "approximation to the sign function using erf(delta*a) ; given delta"

    def generate(
            self,
            degree=7,
            delta=2,
            ensure_bounded=True,
            return_scale=False):
        '''
        Approximation to sign function, using erf(delta * x)
        '''
        degree = int(degree)
        print(f"[pyqsp.poly.PolySign] degree={degree}, delta={delta}")
        if not (degree % 2):
            raise Exception("[PolyErf] degree must be odd")

        def erf_delta(x):
            return scipy.special.erf(x * delta)

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                erf_delta,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)
        else:
            the_poly = self.taylor_series(
                erf_delta,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)

        pcoefs = the_poly.coef
        # force even coefficients to be zero, since the polynomial must be odd
        pcoefs[0::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return TargetPolynomial(pcoefs, target=lambda x: np.sign(x))


class PolyThreshold(PolyTaylorSeries):

    def help(self):
        return "approximation to a thresholding function at threshold 1/2, using linear combination of erf(delta * a); give degree and delta"

    def generate(self,
                 degree=6,
                 delta=2,
                 ensure_bounded=True,
                 return_scale=False):
        '''
        Approximation to threshold function at a=1/2; use a bandpass built from two erf's
        '''
        degree = int(degree)
        print(f"[pyqsp.poly.PolyThreshold] degree={degree}, delta={delta}")
        if (degree % 2):
            raise Exception("[PolyThreshold] degree must be even")

        def erf_delta(x):
            return scipy.special.erf(x * delta)

        def threshold(x):
            return (erf_delta(x + 0.5) - erf_delta(x - 0.5)) / 2

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                threshold,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)
        else:
            the_poly = self.taylor_series(
                threshold,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return pcoefs


class PolyPhaseEstimation(PolyTaylorSeries):

    def help(self):
        return "phase estimation polynomial given "

    def generate(self,
                 degree=6,
                 delta=2,
                 ensure_bounded=True,
                 return_scale=False):
        '''
        Approximation to threshold function at a=1/2; use a bandpass built from two erf's
        '''
        degree = int(degree)
        print(f"[pyqsp.poly.PolyThreshold] degree={degree}, delta={delta}")
        if (degree % 2):
            raise Exception("[PolyThreshold] degree must be even")

        def erf_delta(x):
            return scipy.special.erf(x * delta)

        def threshold(x):
            return (-1 + erf_delta(1/np.sqrt(2) - x) + erf_delta(1/np.sqrt(2) + x))

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                threshold,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)
        else:
            the_poly = self.taylor_series(
                threshold,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return pcoefs


class PolyRect(PolyTaylorSeries):

    def help(self):
        return "approximation to a thresholding function at threshold 1/2, using linear combination of erf(delta * a); give degree and delta"

    def generate(self,
                 degree=6,
                 delta=2,
                 kappa=3,
                 epsilon=0.1,
                 ensure_bounded=True,
                 return_scale=False):
        '''
        Approximation to threshold function at a=1/2; use a bandpass built from two erf's
        '''
        degree = int(degree)
        print(f"[pyqsp.poly.PolyThreshold] degree={degree}, delta={delta}")
        if (degree % 2):
            raise Exception("[PolyThreshold] degree must be even")

        k = np.sqrt(2) / delta * np.sqrt(np.log(2 / (np.pi * epsilon**2)))

        def erf_delta(x):
            return scipy.special.erf(x * k)

        def rect(x):
            return 1 + (erf_delta(x - 3 / (4 * kappa)) +
                        erf_delta(-x - 3 / (4 * kappa))) / 2

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                rect,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)
        else:
            the_poly = self.taylor_series(
                rect,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=0.9)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return pcoefs


# -----------------------------------------------------------------------------


class PolyGibbs(PolyTaylorSeries):
    '''
    exponential decay polynomial
    '''

    def help(self):
        return "approximation to exp(-beta*a) ; specify degree and beta"

    def generate(self,
                 degree=6,
                 beta=2,
                 ensure_bounded=True,
                 return_scale=False):
        degree = int(degree)
        print(f"[pyqsp.poly.PolyGibbs] degree={degree}, beta={beta}")
        if (degree % 2):
            raise Exception("[PolyGibbs] degree must be even")

        def gibbs(x):
            return np.exp(-beta * abs(x))

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                gibbs,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=1)
        else:
            the_poly = self.taylor_series(
                gibbs,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=1)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return TargetPolynomial(pcoefs, target=lambda x: gibbs(x))

# -----------------------------------------------------------------------------


class PolyEigenstateFiltering(PolyTaylorSeries):
    '''
    Lin and Tong's eigenstate filtering polynomial
    '''

    def help(self):
        return "Lin and Tong's eigenstate filtering polynomial ; specify degree, delta, max_scale"

    def generate(
            self,
            degree=6,
            delta=0.2,
            max_scale=0.9,
            ensure_bounded=True,
            return_scale=False):
        degree = int(degree)
        print(f"[pyqsp.poly.PolyEfilter] degree={degree}, delta={delta}")
        if (degree % 2):
            raise Exception("[PolyEfilter] degree must be even")

        def cheb(x):
            Tk = np.polynomial.chebyshev.Chebyshev([0] * degree + [1])
            return Tk(-1 + 2 * (x**2 - delta**2) / (1 - delta**2))
        scale = 1 / cheb(0)

        def efpoly(x):
            return scale * cheb(x)

        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                efpoly,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=max_scale)
        else:
            the_poly, scale = self.taylor_series(
                efpoly,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=max_scale)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return pcoefs

# -----------------------------------------------------------------------------


class PolyRelu(PolyTaylorSeries):
    '''
    Relu function
    '''

    def help(self):
        return "symmetric Relu function sigma(|a-delta|) = 0 if |a| < delta, else |a|-delta ; specify degree, delta"

    def generate(
            self,
            degree=6,
            delta=0.2,
            max_scale=0.99,
            ensure_bounded=True):
        degree = int(degree)
        print(f"[pyqsp.poly.PolyRelu] degree={degree}, delta={delta}")
        if (degree % 2):
            raise Exception("[PolyRelu] degree must be even")

        def cdf(x):
            return (1 + scipy.special.erf(x / np.sqrt(2))) / 2

        def gelu(x):
            return abs(x) * cdf(abs(x) - delta)
        the_poly = self.taylor_series(
            gelu,
            degree,
            ensure_bounded=ensure_bounded,
            max_scale=max_scale)
        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        return pcoefs


class PolySoftPlus(PolyTaylorSeries):
    '''
    SoftPlus function
    '''

    def help(self):
        return "symmetric softplus function sigma(|a-delta|) = 0 if |a| < delta, else |a| ; specify degree, delta"

    def generate(
            self,
            degree=6,
            delta=0.2,
            kappa=1,
            max_scale=0.90,
            ensure_bounded=True,
            return_scale=False):
        degree = int(degree)
        print(
            f"[pyqsp.poly.PolySoftPlus] degree={degree}, delta={delta}, kappa={kappa}")
        if (degree % 2):
            raise Exception("[PolySoftPlus] degree must be even")

        def func(x):
            return np.log(1 + np.exp(kappa * (abs(x) - delta))) / kappa
        if ensure_bounded and return_scale:
            the_poly, scale = self.taylor_series(
                func,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=max_scale)
        else:
            the_poly = self.taylor_series(
                func,
                degree,
                ensure_bounded=ensure_bounded,
                return_scale=return_scale,
                max_scale=max_scale)

        pcoefs = the_poly.coef
        # force odd coefficients to be zero, since the polynomial must be even
        pcoefs[1::2] = 0
        if ensure_bounded and return_scale:
            return pcoefs, scale
        else:
            return pcoefs

# -----------------------------------------------------------------------------


polynomial_generators = {'invert': PolyOneOverX,
                         'poly_sign': PolySign,
                         'poly_thresh': PolyThreshold,
                         'gibbs': PolyGibbs,
                         'efilter': PolyEigenstateFiltering,
                         'relu': PolyRelu,
                         'softplus': PolySoftPlus,
                         }



import numpy as np
from numpy.polynomial.polynomial import Polynomial, polyfromroots
from scipy.special import chebyt, chebyu

from pyqsp.LPoly import Id, LAlg, LPoly


class CompletionError(Exception):
    """Raised when completion step fails."""
    pass


def cheb2poly(ccoefs, kind="T"):
    pcoefs = np.zeros(len(ccoefs), dtype=ccoefs.dtype)
    cfunc = None
    if kind == "T":
        cfunc = chebyt
    elif kind == "U":
        cfunc = chebyu
    else:
        raise Exception("Invalid kind specifier: {}".format(kind))

    for deg, ccoef in enumerate(ccoefs):
        basis = cfunc(deg).coef[::-1]
        pcoefs[:deg + 1] += ccoef * basis

    return pcoefs


def poly2cheb(pcoefs, kind="T"):
    ccoefs = np.zeros(len(pcoefs), dtype=pcoefs.dtype)
    cfunc = None
    if kind == "T":
        cfunc = chebyt
    elif kind == "U":
        cfunc = chebyu
    else:
        raise Exception("Invalid kind specifier: {}".format(kind))

    for i, pcoef in enumerate(pcoefs[::-1]):
        deg = pcoefs.size - i - 1
        basis = cfunc(deg).coef[::-1]
        ccoefs[deg] = pcoef / basis[-1]
        pcoefs[:deg + 1] -= ccoefs[deg] * basis

    return ccoefs


def _pq_completion(P):
    """
    Find polynomial Q given P such that the following matrix is unitary.
    [[P(a), i Q(a) sqrt(1-a^2)],
      i Q*(a) sqrt(1-a^2), P*(a)]]

    Args:
        P: Polynomial object P.

    Returns:
        Polynomial object Q giving described unitary matrix.
    """
    pcoefs = P.coef

    P = Polynomial(pcoefs)
    Pc = Polynomial(pcoefs.conj())
    roots = (1. - P * Pc).roots()

    # catagorize roots
    real_roots = np.array([], dtype=np.float64)
    imag_roots = np.array([], dtype=np.float64)
    cplx_roots = np.array([], dtype=np.complex128)

    tol = 1e-6
    for root in roots:
        if np.abs(np.imag(root)) < tol:
            # real roots
            real_roots = np.append(real_roots, np.real(root))
        elif np.real(root) > -tol and np.imag(root) > -tol:
            if np.real(root) < tol:
                imag_roots = np.append(imag_roots, np.imag(root))
            else:
                cplx_roots = np.append(cplx_roots, root)

    # remove root r_i = +/- 1
    ridx = np.argmin(np.abs(1. - real_roots))
    real_roots = np.delete(real_roots, ridx)
    ridx = np.argmin(np.abs(-1. - real_roots))
    real_roots = np.delete(real_roots, ridx)

    # choose real roots in +/- pairs
    real_roots = np.sort(real_roots)
    real_roots = real_roots[::2]

    # include negative conjugate of complex roots
    cplx_roots = np.r_[cplx_roots, -cplx_roots]

    # construct Q
    Q = Polynomial(
        polyfromroots(
            np.r_[
                real_roots,
                1j *
                imag_roots,
                cplx_roots]))
    Qc = Polynomial(Q.coef.conj())

    # normalize
    norm = np.sqrt((1 - P * Pc).coef[-1] /
                   (Q * Qc * Polynomial([1, 0, -1])).coef[-1])

    return Q * norm


def _fg_completion(F, seed):
    """
    Find polynomial G given Laurent polynomial F such that the following matrix
    is unitary.
    [[F(w), i G(w)],
      i G(1/w), F(1/w)]]

    Args:
        F: Laurent Polynomial object P.
        seed: Random seed.

    Returns:
        Laurent polynomial object G corresponding to a unitary matrix.
    """
    # Find all the roots of Id - (p * ~p) that lies within the upper unit
    # circle.
    poly = (Id - (F * ~F)).coefs
    roots = np.roots(poly)
    # poly is a real, self-inverse Laurent polynomial with no root on the unit circle. All real roots come in reciprocal pairs,
    # and all complex roots come in quadruples (r, r*, 1/r, 1/r*).
    # For each pair of real roots, select the one within the unit circle.
    # For each quadruple of complex roots, select the pair within the unit
    # circle.
    imag_roots = []
    real_roots = []
    for i in roots:
        if (np.abs(i) < 1) and (np.imag(i) > -1e-8):
            if np.imag(i) == 0.:
                real_roots.append(np.real(i))
            else:
                imag_roots.append(i)
    norm = poly[-1]

    # Randomly choose whether to pick the real root (the pair of complex roots)
    #   inside or outside the unit circle.
    # This is to reduce the range of the coefficients appearing in the final
    #   product.
    degree = len(real_roots) + 2 * len(imag_roots)
    lst = []
    if seed is None:
        seed = np.random.randint(2, size=len(imag_roots) + len(real_roots))
    for i, root in enumerate(imag_roots):
        if seed[i]:
            root = 1 / root
        lst.append(LPoly([np.abs(root) ** 2, -2 * np.real(root), 1]))
    for i, root in enumerate(real_roots):
        if seed[i + len(imag_roots)]:
            root = 1 / root
        lst.append(LPoly([-root, 1]))

    # Multiply all the polynomial factors via fft for numerical stability.
    pp = int(np.floor(np.log2(degree))) + 1
    lst_fft = np.pi * np.linspace(0, 1 - 1 / 2**pp, 2**pp)
    coef_mat = np.log(np.array([i.eval(lst_fft) for i in lst]))
    coef_fft = np.exp(np.sum(coef_mat, axis=0))
    gcoefs = np.real(np.fft.fft(coef_fft, 1 << pp))[
        :degree + 1] / (1 << pp)

    # Normalization
    G = LPoly(gcoefs * np.sqrt(norm / gcoefs[0]), -len(gcoefs) + 1)

    return G


def completion_from_root_finding(coefs, coef_type="F", seed=None, tol=1e-6):
    """
    Find a Low Algebra element g such that the identity components are given by
    the input p.

    Args:
        coefs: Array corresponding to the (Laurent) polynomial coefficients.
        coef_type: Quantum signal processing convention in ['F', 'P'].
        seed: Random seed.
        tol: Error tolerance for completion.

    Returns:
        LAlg corresponding to unitary element of the F(w) + G(w) * iX algebra.

    Raises:
        CompletionError: Raised when polynomial cannot be completed.
    """
    if coef_type == "F" or coef_type == "f":
        F = LPoly(coefs, -len(coefs) + 1)
        G = _fg_completion(F, seed)

        ipoly = F
        xpoly = G
    elif coef_type == "P" or coef_type == "p":
        pcoefs = np.array(coefs, dtype=np.complex128)
        P = Polynomial(pcoefs)
        Q = _pq_completion(P)
        deg = pcoefs.size - 1

        pcheb = poly2cheb(pcoefs, kind='T')
        qcheb = np.r_[0., poly2cheb(Q.coef, kind='U')]

        pcheb = pcheb[deg % 2::2]
        qcheb = qcheb[deg % 2::2]

        fcoefs = np.zeros(deg + 1)
        gcoefs = np.zeros(deg + 1)

        if deg % 2 == 0:
            fcoefs[:(deg + 1) // 2] = np.real(pcheb[:0:-1] - qcheb[:0:-1]) / 2
            fcoefs[(deg + 1) // 2 + 1:] = np.real(pcheb[1:] + qcheb[1:]) / 2

            gcoefs[:(deg + 1) // 2] = np.imag(pcheb[:0:-1] + qcheb[:0:-1]) / 2
            gcoefs[(deg + 1) // 2 + 1:] = np.imag(pcheb[1:] - qcheb[1:]) / 2

            fcoefs[(deg + 1) // 2] = np.real(pcheb[0])
            gcoefs[(deg + 1) // 2] = np.imag(pcheb[0])
        else:
            fcoefs[:(deg + 1) // 2] = np.real(pcheb[::-1] - qcheb[::-1]) / 2
            fcoefs[(deg + 1) // 2:] = np.real(pcheb + qcheb) / 2

            gcoefs[:(deg + 1) // 2] = np.imag(pcheb[::-1] + qcheb[::-1]) / 2
            gcoefs[(deg + 1) // 2:] = np.imag(pcheb - qcheb) / 2

        ipoly = LPoly(fcoefs, -len(fcoefs) + 1)
        xpoly = LPoly(gcoefs, -len(gcoefs) + 1)
    else:
        raise CompletionError(
            "Invalid QSP coef_type specifier: {}".format(coef_type))

    # check completion
    ncoefs = (ipoly * ~ipoly + xpoly * ~xpoly).coefs
    ncoefs_expected = np.zeros(ncoefs.size)
    ncoefs_expected[ncoefs.size // 2] = 1.
    success = np.max(np.abs(ncoefs - ncoefs_expected)) < tol

    if not success:
        raise CompletionError(
            "Completion Failed. Input {} = {} could not be completed".format(
                coef_type, coefs))

    return LAlg(ipoly, xpoly)



import glob

from setuptools import setup


def findfiles(pat):
    return [x for x in glob.glob('share/' + pat)]


data_files = [
]

with open("README.md", "r") as fh:
    long_description = fh.read()

# print "data_files = %s" % data_files

setup(
    name='pyqsp',
    version='0.1.4',
    author='I. Chuang, J. Docter, J.M. Martyn, Z. Rossi, A. Tan',
    author_email='ichuang@mit.edu',
    packages=['pyqsp', 'pyqsp.test', 'pyqsp.qsp_models'],
    scripts=[],
    url='https://github.com/ichuang/pyqsp',
    license='LICENSE.txt',
    description='Generate phase angles for quantum signal processing algorithms',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pyqsp = pyqsp.main:CommandLine',
        ],
    },
    install_requires=['matplotlib',
                      'numpy~=1.19.2',
                      'scipy',
                      ],
    package_dir={'pyqsp': 'pyqsp'},
    test_suite="pyqsp.test",
)
